import tests.pgautofailover_utils as pgautofailover
from nose.tools import *

import os
import json

cluster = None
node1 = None
node2 = None
node3 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/no-monitor/node1")
    node1.create(monitorDisabled=True, host=str(node1.vnode.address), nodeId=1)
    node1.run(name="a")


def test_002_init_to_single():
    node1.do_fsm_assign("single")


def test_003_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_004_init_secondary():
    global node2
    node2 = cluster.create_datanode("/tmp/no-monitor/node2")
    node2.create(monitorDisabled=True, host=str(node2.vnode.address), nodeId=2)
    node2.run(name="b")


def test_005_fsm_nodes_set():
    nodesArray = [node1.jsDict("0/1", True), node2.jsDict("0/1", False)]

    node1.do_fsm_nodes_set(nodesArray)
    node2.do_fsm_nodes_set(nodesArray)


def test_006_init_to_wait_standby():
    node2.do_fsm_assign("wait_standby")


def test_007_catchingup():
    node1.do_fsm_assign("wait_primary")
    node2.do_fsm_assign("catchingup")


def test_008_secondary():
    node1.do_fsm_assign("primary")
    node2.do_fsm_assign("secondary")

    eq_(node1.get_synchronous_standby_names_local(), "*")


def test_009_init_secondary():
    global node3
    node3 = cluster.create_datanode("/tmp/no-monitor/node3")
    node3.create(monitorDisabled=True, host=str(node3.vnode.address), nodeId=3)
    node3.run(name="c")


def test_010_fsm_nodes_set():
    LSN1 = node1.run_sql_query("select pg_current_wal_flush_lsn()")[0][0]
    LSN2 = node2.run_sql_query("select pg_last_wal_receive_lsn()")[0][0]
    nodesArray = [
        node1.jsDict(LSN1, True),
        node2.jsDict(LSN2, False),
        node3.jsDict("0/1", False),
    ]

    node1.do_fsm_nodes_set(nodesArray)
    node2.do_fsm_nodes_set(nodesArray)
    node3.do_fsm_nodes_set(nodesArray)


def test_011_init_to_wait_standby():
    node1.do_fsm_assign("primary")
    node3.do_fsm_assign("wait_standby")

    eq_(node1.get_synchronous_standby_names_local(), "*")


def test_012_catchingup():
    node3.do_fsm_assign("catchingup")

    eq_(node1.get_synchronous_standby_names_local(), "*")


def test_013_secondary():
    node3.do_fsm_assign("secondary")
    node1.do_fsm_assign("primary")

    # no monitor: use the generic value '*'
    eq_(node1.get_synchronous_standby_names_local(), "*")



import tests.pgautofailover_utils as pgautofailover
from nose.tools import *

cluster = None
node1 = None
node2 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    monitor = cluster.create_monitor("/tmp/listen/monitor")
    monitor.run()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/listen/node1", listen_flag=True)
    node1.create()
    node1.run()
    assert node1.wait_until_state(target_state="single")
    node1.wait_until_pg_is_running()


def test_002_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_init_secondary():
    global node2
    node2 = cluster.create_datanode("/tmp/listen/node2", listen_flag=True)
    node2.create()
    node2.run()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


def test_004_read_from_secondary():
    results = node2.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,)]


@raises(Exception)
def test_005_writes_to_node2_fail():
    node2.run_sql_query("INSERT INTO t1 VALUES (3)")


def test_006_fail_primary():
    node1.fail()
    assert node2.wait_until_state(target_state="wait_primary")


def test_007_writes_to_node2_succeed():
    node2.run_sql_query("INSERT INTO t1 VALUES (3)")
    results = node2.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,), (3,)]


def test_008_start_node1_again():
    node1.run()
    assert node2.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")


def test_009_read_from_new_secondary():
    results = node1.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,), (3,)]


@raises(Exception)
def test_010_writes_to_node1_fail():
    node1.run_sql_query("INSERT INTO t1 VALUES (3)")


def test_011_fail_secondary():
    node1.fail()
    assert node2.wait_until_state(target_state="wait_primary")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# pg_auto_failover documentation build configuration file, created by
# sphinx-quickstart on Sat May  5 14:33:23 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

#
# https://stackoverflow.com/questions/9899283/how-do-you-change-the-code-example-font-size-in-latex-pdf-output-with-sphinx
#

from sphinx.highlighting import PygmentsBridge
from pygments.formatters.latex import LatexFormatter


class CustomLatexFormatter(LatexFormatter):
    def __init__(self, **options):
        super(CustomLatexFormatter, self).__init__(**options)
        self.verboptions = r"formatcom=\tiny"


PygmentsBridge.latex_formatter = CustomLatexFormatter


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.githubpages"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "pg_auto_failover"
copyright = "Copyright (c) Microsoft Corporation. All rights reserved."
author = "Microsoft"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "2.1"
# The full version, including alpha/beta/rc tags.
release = "2.1"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".venv"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}


# Add our custom CSS
def setup(app):
    if hasattr(app, "add_css_file"):
        app.add_css_file("css/citus.css")
        app.add_css_file("css/pygments.css")


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "pg_auto_failoverdoc"


# -- Options for LaTeX output ---------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "pg_auto_failover.tex",
        "pg\_auto\_failover Documentation",
        "Microsoft",
        "manual",
    ),
]

# for some reasons, doesn't work with the current setup.
# latex_engine = 'xelatex'

latex_elements = {
    "geometry": r"\usepackage[paperwidth=441pt,paperheight=666pt]{geometry}",
    "pointsize": "12pt",
    "fncychap": r"\usepackage[Sonny]{fncychap}",
    "extraclassoptions": "oneside",
    "sphinxsetup": "hmargin=1cm,vmargin=2cm,verbatimwithframe=false,VerbatimColor={rgb}{1.0, 0.97, 0.97}",
}

latex_show_urls = "footnote"

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "pg_auto_failover",
        "pg_auto_failover Documentation",
        [author],
        1,
    ),
    ("ref/pg_autoctl", "pg_autoctl", "pg_autoctl", [author], 1),
    (
        "ref/pg_autoctl_create_monitor",
        "pg_autoctl create monitor",
        "pg_autoctl create monitor",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_create_postgres",
        "pg_autoctl create postgres",
        "pg_autoctl create postgres",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_create_formation",
        "pg_autoctl create formation",
        "pg_autoctl create formation",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_drop_monitor",
        "pg_autoctl drop monitor",
        "pg_autoctl drop monitor",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_drop_node",
        "pg_autoctl drop node",
        "pg_autoctl drop node",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_drop_formation",
        "pg_autoctl drop formation",
        "pg_autoctl drop formation",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_config_get",
        "pg_autoctl config get",
        "pg_autoctl config get",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_config_set",
        "pg_autoctl config set",
        "pg_autoctl config set",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_config_check",
        "pg_autoctl config check",
        "pg_autoctl config check",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_show_uri",
        "pg_autoctl show uri",
        "pg_autoctl show uri",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_show_events",
        "pg_autoctl show events",
        "pg_autoctl show events",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_show_state",
        "pg_autoctl show state",
        "pg_autoctl show state",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_show_file",
        "pg_autoctl show file",
        "pg_autoctl show file",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_show_settings",
        "pg_autoctl show settings",
        "pg_autoctl show settings",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_show_standby_names",
        "pg_autoctl show standby-names",
        "pg_autoctl show standby-names",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_enable_maintenance",
        "pg_autoctl enable maintenance",
        "pg_autoctl enable maintenance",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_enable_secondary",
        "pg_autoctl enable secondary",
        "pg_autoctl enable secondary",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_enable_ssl",
        "pg_autoctl enable ssl",
        "pg_autoctl enable ssl",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_enable_monitor",
        "pg_autoctl enable monitor",
        "pg_autoctl enable monitor",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_disable_maintenance",
        "pg_autoctl disable maintenance",
        "pg_autoctl disable maintenance",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_disable_secondary",
        "pg_autoctl disable secondary",
        "pg_autoctl disable secondary",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_disable_ssl",
        "pg_autoctl disable ssl",
        "pg_autoctl disable ssl",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_disable_monitor",
        "pg_autoctl disable monitor",
        "pg_autoctl disable monitor",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_get_formation_settings",
        "pg_autoctl get formation settings",
        "pg_autoctl get formation settings",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_get_formation_number_sync_standbys",
        "pg_autoctl get formation number-sync-standbys",
        "pg_autoctl get formation number-sync-standbys",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_get_node_replication_quorum",
        "pg_autoctl get node replication-quorum",
        "pg_autoctl get node replication-quorum",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_get_node_candidate_priority",
        "pg_autoctl get node candidate-priority",
        "pg_autoctl get node candidate-priority",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_set_formation_number_sync_standbys",
        "pg_autoctl set formation number-sync-standbys",
        "pg_autoctl set formation number-sync-standbys",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_set_node_replication_quorum",
        "pg_autoctl set node replication-quorum",
        "pg_autoctl set node replication-quorum",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_set_node_candidate_priority",
        "pg_autoctl set node candidate-priority",
        "pg_autoctl set node candidate-priority",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_perform_failover",
        "pg_autoctl perform failover",
        "pg_autoctl perform failover",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_perform_switchover",
        "pg_autoctl perform switchover",
        "pg_autoctl perform switchover",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_perform_promotion",
        "pg_autoctl perform promotion",
        "pg_autoctl perform promotion",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_run",
        "pg_autoctl run",
        "pg_autoctl run",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_watch",
        "pg_autoctl watch",
        "pg_autoctl watch",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_stop",
        "pg_autoctl stop",
        "pg_autoctl stop",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_status",
        "pg_autoctl status",
        "pg_autoctl status",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_reload",
        "pg_autoctl reload",
        "pg_autoctl reload",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_create_coordinator",
        "pg_autoctl create coordinator",
        "pg_autoctl create coordinator",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_create_worker",
        "pg_autoctl create worker",
        "pg_autoctl create worker",
        [author],
        1,
    ),
    (
        "ref/pg_autoctl_activate",
        "pg_autoctl activate",
        "pg_autoctl activate",
        [author],
        1,
    ),
    # ("ref/reference", "pg_autoctl", "pg_auto_failover agent", [author], 1),
    (
        "ref/configuration",
        "pg_autoctl",
        "pg_auto_failover Configuration",
        [author],
        5,
    ),
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "pg_auto_failover",
        "pg_auto_failover Documentation",
        author,
        "pg_auto_failover",
        "One line description of project.",
        "Miscellaneous",
    ),
]



import tests.pgautofailover_utils as pgautofailover
from nose.tools import raises, eq_
import time
import subprocess

import os.path

cluster = None
monitor = None
node1 = None
node2 = None
node3 = None
node4 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    global monitor

    # test creating the monitor in an existing empty directory
    p = subprocess.Popen(
        [
            "sudo",
            "-E",
            "-u",
            os.getenv("USER"),
            "env",
            "PATH=" + os.getenv("PATH"),
            "mkdir",
            "-p",
            "/tmp/multi_async/monitor",
        ]
    )
    assert p.wait() == 0

    monitor = cluster.create_monitor("/tmp/multi_async/monitor")
    monitor.run()
    monitor.wait_until_pg_is_running()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/multi_async/node1")
    node1.create()
    node1.run()
    assert node1.wait_until_state(target_state="single")


def test_002_add_standby():
    global node2

    node2 = cluster.create_datanode("/tmp/multi_async/node2")
    node2.create()
    node2.run()

    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()

    # make sure we reached primary on node1 before next tests
    assert node1.wait_until_state(target_state="primary")


def test_003_add_standby():
    global node3

    node3 = cluster.create_datanode("/tmp/multi_async/node3")
    node3.create(level="-vv", replicationQuorum=False)
    node3.run()

    assert node3.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()
    assert node3.has_needed_replication_slots()

    # the formation number_sync_standbys is expected to still be zero now
    eq_(node1.get_number_sync_standbys(), 0)

    # make sure we reached primary on node1 before next tests
    assert node1.wait_until_state(target_state="primary")


def test_004_set_async():
    # now we set the whole formation to async
    node1.set_number_sync_standbys(0)
    eq_(node1.get_number_sync_standbys(), 0)

    print()
    assert node1.wait_until_state(target_state="primary")

    assert node1.set_replication_quorum("false")  # primary
    assert node2.set_replication_quorum("false")  # secondary
    assert node3.set_replication_quorum("false")  # secondary


def test_005_write_into_primary():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2), (3), (4)")
    node1.run_sql_query("CHECKPOINT")

    results = node1.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,), (3,), (4,)]


def test_006_async_failover():
    print()
    print("Calling pgautofailover.failover() on the monitor")
    node2.perform_promotion()

    assert node1.wait_until_state(target_state="secondary")  # secondary
    assert node3.wait_until_state(target_state="secondary")  # secondary
    assert node2.wait_until_state(target_state="primary")  # primary


def test_007_read_from_new_primary():
    results = node2.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,), (3,), (4,)]


#
# The next tests prepare a test-case where at promotion time an async
# standby is first driven to SECONDARY, and then other sync standby nodes in
# REPORT_LSN move forward. We had a bug where the REPORT_LSN nodes would be
# stuck with the primary node being in the WAIT_PRIMARY/PRIMARY state.
#
def test_008_set_sync_async():
    print()

    assert node1.set_replication_quorum("true")  # secondary
    assert node2.set_replication_quorum("true")  # primary
    assert node3.set_replication_quorum("false")  # secondary

    assert node3.set_candidate_priority(0)

    assert node2.wait_until_state(target_state="primary")


def test_009_add_sync_standby():
    global node4

    node4 = cluster.create_datanode("/tmp/multi_async/node4")
    node4.create()
    node4.run()

    assert node1.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="primary")
    assert node3.wait_until_state(target_state="secondary")
    assert node4.wait_until_state(target_state="secondary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()
    assert node3.has_needed_replication_slots()
    assert node4.has_needed_replication_slots()

    # the formation number_sync_standbys is expected to be incremented, we
    # now have two standby nodes that participate in the replication quorum
    # (node1 and node4)
    eq_(node2.get_number_sync_standbys(), 1)

    # make sure we reached primary on node1 before next tests
    assert node2.wait_until_state(target_state="primary")


def test_010_promote_node1():
    print()
    print("Calling pgautofailover.perform_promotion(node1) on the monitor")

    # we don't use node1.perform_promotion() here because using the
    # pg_autoctl client means we would listen to notification and get back
    # to the rest of the code when the promotion is all over with
    #
    # we need to take control way before that, so just trigger the failover
    # and get back to controlling our test case.
    q = "select pgautofailover.perform_promotion('default', 'node_1')"
    monitor.run_sql_query(q)


def test_011_ifdown_node4_at_reportlsn():
    print()
    assert node4.wait_until_state(target_state="report_lsn")
    node4.ifdown()

    assert node3.wait_until_state(target_state="secondary")


def test_012_ifup_node4():
    node4.ifup()

    print()
    assert node3.wait_until_state(target_state="secondary")
    assert node4.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")
    assert node2.wait_until_state(target_state="secondary")


def test_013_drop_node4():
    node4.destroy()

    print()
    assert node1.wait_until_state(target_state="primary")
    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="secondary")

    # we have only one standby participating in the quorum now
    eq_(node1.get_number_sync_standbys(), 0)


#
# A series of test where we fail primary and candidate secondary node and
# all is left is a secondary that is not a candidate for failover.
#
# In the first series 014_0xx the demoted primary comes back first and needs
# to fetch missing LSNs from node3 because it might have missed some
# transactions.
#
#   node1         node2        node3
#   primary       secondary    secondary
#   <down>
#   demoted       wait_primary secondary
#                 <down>
#   demoted       demoted      report_lsn
#   <up>
#   wait_primary  demoted      secondary   (fast_forward → primary)
#                 <up>
#   primary       secondary    secondary
#
def test_014_001_fail_node1():
    node1.fail()

    # first we have a 30s timeout for the monitor to decide that node1 is
    # down; then we have another 30s timeout at stop_replication waiting for
    # demote_timeout, so let's give it 120s there and step in the middle first
    assert node2.wait_until_state(target_state="stop_replication", timeout=120)
    assert node2.wait_until_state(target_state="wait_primary", timeout=120)
    assert node3.wait_until_state(target_state="secondary")

    node2.check_synchronous_standby_names(ssn="")


def test_014_002_stop_new_primary_node2():
    node2.fail()

    print()
    assert node3.wait_until_state(target_state="report_lsn")


def test_014_003_restart_node1():
    node1.run()

    # node1 used to be primary, now demoted, and meanwhile node2 was primary
    # node1 is assigned report_lsn and then is selected (only node with
    # candidate priority > 0) ; and thus needs to go through fast_forward
    assert node1.wait_until_assigned_state(target_state="fast_forward")
    assert node1.wait_until_state(target_state="stop_replication")
    assert node1.wait_until_state(target_state="wait_primary")
    assert node3.wait_until_state(target_state="secondary")


def test_014_004_restart_node2():
    node2.run()

    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


#
# Test 15 is like test 14, though we inverse the restarting of the failed
# nodes, first the new primary and then very old primary.
#
#   node1        node2        node3
#   primary      secondary    secondary
#   <down>
#   demoted      wait_primary secondary
#                <down>
#   demoted      demoted      report_lsn
#                <up>
#   demoted      wait_primary secondary
#   <up>
#   secondary    primary      secondary
#
def test_015_001_fail_primary_node1():
    node1.fail()

    # first we have a 30s timeout for the monitor to decide that node1 is
    # down; then we have another 30s timeout at stop_replication waiting for
    # demote_timout, so let's give it 120s there
    assert node2.wait_until_state(target_state="wait_primary", timeout=120)
    assert node3.wait_until_state(target_state="secondary")

    node2.check_synchronous_standby_names(ssn="")


def test_015_002_fail_new_primary_node2():
    node2.fail()

    print()
    assert node3.wait_until_state(target_state="report_lsn")


def test_015_003_restart_node2():
    node2.run()

    # restart the previous primary, it re-joins as a (wannabe) primary
    # because the only secondary has candidatePriority = 0, it's wait_primary
    assert node2.wait_until_state(target_state="wait_primary")
    assert node3.wait_until_state(target_state="secondary")

    time.sleep(5)
    assert not node2.get_state().assigned == "primary"


def test_015_004_restart_node1():
    node1.run()

    assert node3.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")


#
# When after loosing both secondary nodes, the one that's back online has
# candidate priority set to zero, then we should remain in wait_primary
# state.
#
def test_016_001_fail_node3():
    node3.fail()
    assert node3.wait_until_assigned_state(target_state="catchingup")


def test_016_002_fail_node1():
    node1.fail()
    assert node2.wait_until_state(target_state="wait_primary")


def test_016_003_restart_node3():
    node3.run()

    assert node3.wait_until_assigned_state(target_state="secondary")
    assert node2.wait_until_state(target_state="wait_primary")

    time.sleep(5)
    assert not node2.get_state().assigned == "primary"


def test_016_004_restart_node1():
    node1.run()

    assert node3.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")


#
# When a node with candidate-priority zero (here, node3) fails while the
# primary node (here, node2) is already in wait_primary, the non-candidate
# node (here, node3) should still be assigned catchingup.
#
def test_017_001_fail_node1():
    node1.fail()
    assert node2.wait_until_state(target_state="wait_primary")


def test_017_002_fail_node3():
    node3.fail()
    assert node3.wait_until_assigned_state(target_state="catchingup")


def test_017_003_restart_nodes():
    node3.run()
    node1.run()

    assert node3.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")



from nose.tools import *
from tests.tablespaces.tablespace_utils import node1
from tests.tablespaces.tablespace_utils import node2


# Node 1 is brought back online by makefile, wait until it's up and replicating
def test_001_original_primary_comes_back_up():
    assert node1.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="primary")

    results = node1.run_sql_query("SELECT * FROM t4")
    assert results == [(10,), (11,)]
    results = node1.run_sql_query("SELECT * FROM t2")
    assert results == [(3,), (4,), (7,), (12,)]
    results = node1.run_sql_query("SELECT * FROM t3")
    assert results == [(5,), (6,), (8,), (13,)]



import tests.pgautofailover_utils as pgautofailover
from nose.tools import raises, eq_
import time

import os.path

cluster = None
monitor = None
node1 = None
node2 = None
node3 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    global monitor
    monitor = cluster.create_monitor(
        "/tmp/multi_alternate_primary_failures/monitor"
    )
    monitor.run()
    monitor.wait_until_pg_is_running()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode(
        "/tmp/multi_alternate_primary_failures/node1"
    )
    node1.create()
    node1.run()
    print()
    assert node1.wait_until_state(target_state="single")


def test_002_001_add_two_standbys():
    global node2

    node2 = cluster.create_datanode(
        "/tmp/multi_alternate_primary_failures/node2"
    )
    node2.create()
    node2.run()

    node2.wait_until_pg_is_running()

    print()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()

    # with one standby, we have number_sync_standbys set to 0 still
    assert node1.get_number_sync_standbys() == 0


def test_002_002_add_two_standbys():
    global node3

    node3 = cluster.create_datanode(
        "/tmp/multi_alternate_primary_failures/node3"
    )
    node3.create()
    node3.run()

    node3.wait_until_pg_is_running()

    print()
    assert node3.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()
    assert node3.has_needed_replication_slots()

    # with two standbys, we have number_sync_standbys set to 1
    assert node1.get_number_sync_standbys() == 1


#
# In this test series, we have
#
#   node1         node2        node3
#   primary       secondary    secondary
#   <down>
#   demoted       primary      secondary
#                 <down>
#   demoted       draining     report_lsn
#                 <up>
#   demoted       primary      secondary
#   <up>
#   secondary     primary      secondary
#
def test_003_001_stop_primary():
    # verify that node1 is primary and stop it
    assert node1.get_state().assigned == "primary"
    node1.fail()

    # wait for node2 to become the new primary
    print()
    assert node1.wait_until_assigned_state(target_state="demoted")
    assert node2.wait_until_state(target_state="primary")


def test_003_002_stop_primary():
    # verify that node2 is primary and stop it
    assert node2.get_state().assigned == "primary"
    node2.fail()

    # node3 can't be promoted when it's the only one reporting its LSN
    print()
    assert node2.wait_until_assigned_state(target_state="draining")
    assert node3.wait_until_state(target_state="report_lsn")

    # check that node3 stays at report_lsn and doesn't go to wait_primary
    node3.sleep(5)
    assert node3.wait_until_state(target_state="report_lsn")


def test_003_003_bringup_last_failed_primary():
    # Restart node2
    node2.run()

    # Now node 2 should become primary
    print()
    assert node2.wait_until_state(target_state="primary")
    assert node3.wait_until_state(target_state="secondary")


def test_003_004_bringup_first_failed_primary():
    # Restart node1
    node1.run()
    node3.wait_until_pg_is_running()

    # Now node 1 should become secondary
    print()
    assert node1.wait_until_state(target_state="secondary")
    assert node2.get_state().assigned == "primary"
    assert node3.get_state().assigned == "secondary"

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()
    assert node3.has_needed_replication_slots()


#
# In this test series , we have
#
#   node1         node2        node3
#   secondary     primary      secondary
#                 <down>
#   primary       demoted      secondary
#   <down>
#   draining      demoted      report_lsn
#                 <up>
#   draining      secondary    primary
#   <up>
#   secondary     secondary    primary
#
def test_005_001_fail_primary_again():
    # verify that node2 is primary and stop it
    assert node2.get_state().assigned == "primary"
    node2.fail()

    print()
    assert node2.wait_until_assigned_state(
        target_state="demote_timeout", timeout=120
    )
    assert node2.wait_until_assigned_state(target_state="demoted", timeout=120)
    assert node1.wait_until_assigned_state(target_state="primary", timeout=120)
    assert node1.wait_until_state(target_state="primary", timeout=120)
    assert node3.wait_until_state(target_state="secondary", timeout=120)


def test_005_002_fail_primary_again():
    # verify that node1 is primary and stop it
    assert node1.get_state().assigned == "primary"
    node1.fail()

    print()
    assert node1.wait_until_assigned_state(target_state="draining")
    assert node3.wait_until_assigned_state(target_state="report_lsn")


def test_005_003_bring_up_first_failed_primary():
    # Restart node2
    node2.run()

    print()
    assert node2.wait_until_state(target_state="demoted")

    # Now node 2 should become secondary
    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="primary")


def test_005_004_bring_up_last_failed_primary():
    # Restart node1
    node1.run()
    node1.wait_until_pg_is_running()

    # Now node 3 should become secondary
    print()
    assert node1.wait_until_state(target_state="secondary")
    assert node3.get_state().assigned == "primary"
    assert node2.get_state().assigned == "secondary"


#
# In this test series , we have
#
#   node1         node2        node3
#   secondary     secondary    primary
#                              <down>
#   primary       secondary    demoted
#   <down>
#                              <up>
#   demoted       primary      secondary
#   <up>
#   secondary     primary      secondary
#
def test_006_001_fail_primary():
    assert node3.get_state().assigned == "primary"
    node3.fail()

    print()
    assert node3.wait_until_assigned_state(
        target_state="demote_timeout", timeout=120
    )
    assert node3.wait_until_assigned_state(target_state="demoted", timeout=120)
    assert node1.wait_until_assigned_state(target_state="primary", timeout=120)
    assert node1.wait_until_state(target_state="primary", timeout=120)
    assert node2.wait_until_state(target_state="secondary", timeout=120)


def test_006_002_fail_new_primary():
    assert node1.get_state().assigned == "primary"
    node1.fail()
    node3.run()

    print()
    assert node2.wait_until_state(target_state="primary", timeout=120)
    assert node3.wait_until_state(target_state="secondary", timeout=120)


def test_006_003_bringup_last_failed_primary():
    node1.run()

    print()
    assert node1.wait_until_state(target_state="secondary", timeout=120)
    assert node2.wait_until_state(target_state="primary", timeout=120)
    assert node3.wait_until_state(target_state="secondary", timeout=120)



#! /usr/bin/env python3

#
# Write a Python application that knows how to exit gracefully when
# receiving SIGTERM (at docker-compose down time), but doesn't know how to
# do much else.
#

import sys
import time
import signal
from datetime import datetime


def sigterm_handler(_signo, _stack_frame):
    sys.exit(0)


signal.signal(signal.SIGTERM, sigterm_handler)

if __name__ == "__main__":
    try:
        while True:
            print("%s" % datetime.now())
            time.sleep(600)
    except BaseException:
        # Keyboard Interrupt or something
        pass



import os
import time

import tests.pgautofailover_utils as pgautofailover
from nose.tools import eq_

cluster = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.monitor.stop_pg_autoctl()
    cluster.destroy()


def test_000_create_monitor():
    monitor = cluster.create_monitor("/tmp/update/monitor")


def test_001_update_extension():
    os.environ["PG_AUTOCTL_DEBUG"] = "1"
    os.environ["PG_AUTOCTL_EXTENSION_VERSION"] = "dummy"

    cluster.monitor.run()

    cluster.monitor.wait_until_pg_is_running()

    # Wait until extension is installed
    time.sleep(1)

    results = cluster.monitor.run_sql_query(
        """SELECT installed_version
             FROM pg_available_extensions
            WHERE name = 'pgautofailover'
        """
    )

    if results[0][0] != "dummy":
        cluster.monitor.print_debug_logs()

    eq_(results, [("dummy",)])

    del os.environ["PG_AUTOCTL_EXTENSION_VERSION"]
    assert "PG_AUTOCTL_EXTENSION_VERSION" not in os.environ



import tests.pgautofailover_utils as pgautofailover

from nose.tools import eq_

import os.path

cluster = None
node1 = None
node2 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    monitor = cluster.create_monitor(
        "/tmp/ssl-self-signed/monitor", sslSelfSigned=True
    )

    monitor.run()
    monitor.wait_until_pg_is_running()
    monitor.check_ssl("on", "require")


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode(
        "/tmp/ssl-self-signed/node1", sslSelfSigned=True
    )
    node1.create()
    node1.run()
    assert node1.wait_until_state(target_state="single")

    node1.wait_until_pg_is_running()
    node1.check_ssl("on", "require", primary=True)


def test_002_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_init_secondary():
    global node2
    node2 = cluster.create_datanode(
        "/tmp/ssl-self-signed/node2", sslSelfSigned=True, sslMode="require"
    )

    node2.create()
    node2.run()

    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    node2.wait_until_pg_is_running()
    node2.check_ssl("on", "require")


def test_004_failover():
    print()
    print("Calling pgautofailover.failover() on the monitor")
    cluster.monitor.failover()
    assert node2.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")



import tests.pgautofailover_utils as pgautofailover
from nose.tools import raises, eq_

import subprocess
import os, os.path, time, shutil
import pprint

cluster = None
coord0a = None
coord0b = None
worker1a = None
worker1b = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    monitor = cluster.create_monitor(
        "/tmp/citus/skip/monitor", authMethod="skip"
    )
    monitor.run()
    monitor.wait_until_pg_is_running()

    with open(os.path.join(monitor.datadir, "pg_hba.conf"), "a") as hba:
        hba.write("host all all %s trust\n" % cluster.networkSubnet)

    monitor.reload_postgres()


def test_001a_init_coordinator():
    global coord0a

    # allow using unix domain sockets
    pgautofailover.sudo_mkdir_p("/tmp/socks/coord0a")
    os.environ["PG_REGRESS_SOCK_DIR"] = "/tmp/socks/coord0a"

    coord0a = cluster.create_datanode(
        "/tmp/citus/skip/coord0a",
        role=pgautofailover.Role.Coordinator,
        authMethod="skip",
    )
    coord0a.create()

    with open(os.path.join(coord0a.datadir, "pg_hba.conf"), "a") as hba:
        # run_sql_query() will need
        # host "172.27.1.1", user "docker", database "postgres"
        hba.write("host postgres docker %s trust\n" % cluster.networkSubnet)
        hba.write("host citus docker %s trust\n" % cluster.networkSubnet)
        hba.write("host all all %s trust\n" % cluster.networkSubnet)
        hba.write("host replication all %s trust\n" % cluster.networkSubnet)

    coord0a.run()
    assert coord0a.wait_until_state(target_state="single")


def test_001b_init_coordinator():
    global coord0b

    # allow using unix domain sockets
    pgautofailover.sudo_mkdir_p("/tmp/socks/coord0b")
    os.environ["PG_REGRESS_SOCK_DIR"] = "/tmp/socks/coord0b"

    # coord0b will fetch the HBA from coord0a during pg_basebackup
    coord0b = cluster.create_datanode(
        "/tmp/citus/skip/coord0b",
        role=pgautofailover.Role.Coordinator,
        authMethod="skip",
    )
    coord0b.create()
    coord0b.run()

    assert coord0a.wait_until_state(target_state="primary")
    assert coord0b.wait_until_state(target_state="secondary")


def test_002a_init_worker():
    global worker1a

    # allow using unix domain sockets
    pgautofailover.sudo_mkdir_p("/tmp/socks/worker1a")
    os.environ["PG_REGRESS_SOCK_DIR"] = "/tmp/socks/worker1a"

    worker1a = cluster.create_datanode(
        "/tmp/citus/skip/worker1a",
        group=1,
        role=pgautofailover.Role.Worker,
        authMethod="skip",
    )


@raises(Exception)
def test_002b_create_worker():
    # we expect failure when calling master_activate_node on the coordinator
    # as the coordinator is not allowed to connect to this node yet
    worker1a.create(name="worker1a")


def test_002c_check_local_state():
    state = worker1a.get_local_state()
    eq_(("init", "single"), state)


def test_002d_run_worker():
    # edit the HBA, allowing the activation retry to succeed this time
    with open(os.path.join(worker1a.datadir, "pg_hba.conf"), "a") as hba:
        hba.write("host all all %s trust\n" % cluster.networkSubnet)
        hba.write("host replication all %s trust\n" % cluster.networkSubnet)

    worker1a.run()


def test_002e_check_hba():
    # check the HBA has not been tampered by pg_autoctl
    eq_(False, worker1a.editedHBA())


def test_002f_activated_node():
    timeout = 10
    q = "select isactive from pg_dist_node where nodename = %s"
    isactive = coord0a.run_sql_query(q, str(worker1a.vnode.address))[0][0]

    while timeout > 0 and isactive is not True:
        print(".", end="")
        time.sleep(1)
        timeout -= 1

        isactive = coord0a.run_sql_query(q, str(worker1a.vnode.address))[0][0]

    if isactive is not True:
        raise Exception(
            "test failed: worker1a is still not activated "
            + "in the coordinator after 10s"
        )


def test_003_init_worker():
    global worker1b

    # allow using unix domain sockets
    pgautofailover.sudo_mkdir_p("/tmp/socks/worker1b")
    os.environ["PG_REGRESS_SOCK_DIR"] = "/tmp/socks/worker1b"

    worker1b = cluster.create_datanode(
        "/tmp/citus/skip/worker1b",
        group=1,
        role=pgautofailover.Role.Worker,
        authMethod="skip",
    )
    worker1b.create(name="worker1b")
    worker1b.run()

    eq_(False, worker1b.editedHBA())

    # wait here till all workers are stable and in the desired state
    # by not waiting on all workers separately this should save a bit of time
    print()  # make the debug output more readable
    assert worker1a.wait_until_state(target_state="primary")
    assert worker1b.wait_until_state(target_state="secondary")


def test_004_create_distributed_table():
    coord0a.run_sql_query("CREATE TABLE t1 (a int)")
    coord0a.run_sql_query("SELECT create_distributed_table('t1', 'a')")
    coord0a.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_005_failover():
    cluster.monitor.failover(group=1)
    assert worker1b.wait_until_state(target_state="wait_primary")
    assert worker1a.wait_until_state(target_state="secondary")
    assert worker1b.wait_until_state(target_state="primary")



import tests.pgautofailover_utils as pgautofailover
from nose.tools import eq_

import subprocess
import os, os.path, time, shutil

cluster = None
node1 = None
node2 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    monitor = cluster.create_monitor("/tmp/skip/monitor", authMethod="skip")
    monitor.run()
    monitor.wait_until_pg_is_running()

    with open(os.path.join("/tmp/skip/monitor", "pg_hba.conf"), "a") as hba:
        hba.write("host all all %s trust\n" % cluster.networkSubnet)

    monitor.reload_postgres()


def test_001_init_primary():
    global node1

    print()
    node1_path = cluster.pg_createcluster("node1")

    # make a copy of the debian's HBA file
    hba_path = os.path.join(
        "/etc", "/".join(node1_path.split("/")[3:]), "pg_hba.conf"
    )
    shutil.copyfile(hba_path, "/tmp/pg_hba.debian.conf")

    # allow using unix domain sockets
    pgautofailover.sudo_mkdir_p("/tmp/socks/node1")
    os.environ["PG_REGRESS_SOCK_DIR"] = "/tmp/socks/node1"

    # we need to give the hostname here, because our method to find it
    # automatically will fail in the test environment
    node1 = cluster.create_datanode(node1_path, authMethod="skip")
    node1.create(level="-vvv")

    #
    # Check that we didn't edit the HBA file, thanks to --skip-pg-hba, here
    # in the test file spelled the strange way --auth skip.
    #
    p = subprocess.run(
        [
            "diff",
            "/tmp/pg_hba.debian.conf",
            os.path.join(node1_path, "pg_hba.conf"),
        ],
        text=True,
        capture_output=True,
    )

    print("diff %s" % " ".join(p.args))

    if p.returncode != 0:
        print("%s" % p.stdout)

    assert p.returncode == 0

    with open(os.path.join(node1_path, "pg_hba.conf"), "a") as hba:
        # node1.run_sql_query will need
        # host "172.27.1.1", user "docker", database "postgres"
        hba.write("host postgres docker %s trust\n" % cluster.networkSubnet)
        hba.write("host all all %s trust\n" % cluster.networkSubnet)
        hba.write("host replication all %s trust\n" % cluster.networkSubnet)

    node1.reload_postgres()

    node1.run()
    assert node1.wait_until_state(target_state="single")
    node1.wait_until_pg_is_running()


def test_002_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_init_secondary():
    global node2

    # allow using unix domain sockets
    pgautofailover.sudo_mkdir_p("/tmp/socks/node2")
    os.environ["PG_REGRESS_SOCK_DIR"] = "/tmp/socks/node2"

    node2 = cluster.create_datanode("/tmp/skip/node2", authMethod="skip")
    node2.create()

    with open(os.path.join("/tmp/skip/node2", "pg_hba.conf"), "a") as hba:
        hba.write("host all all %s trust\n" % cluster.networkSubnet)
        hba.write("host replication all %s trust\n" % cluster.networkSubnet)

    node2.reload_postgres()

    node2.run()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


def test_004a_hba_have_not_been_edited():
    eq_(False, node1.editedHBA())
    eq_(False, node2.editedHBA())


def test_004b_pg_autoctl_conf_skip_hba():
    eq_("skip", node1.config_get("postgresql.hba_level"))
    eq_("skip", node2.config_get("postgresql.hba_level"))


def test_005_failover():
    print()
    print("Calling pgautofailover.failover() on the monitor")
    cluster.monitor.failover()
    assert node2.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")


def test_006_restart_secondary():
    node1.stop_pg_autoctl()
    node1.run()


def test_006a_hba_have_not_been_edited():
    eq_(False, node1.editedHBA())
    eq_(False, node2.editedHBA())


def test_006b_pg_autoctl_conf_skip_hba():
    eq_("skip", node1.config_get("postgresql.hba_level"))
    eq_("skip", node2.config_get("postgresql.hba_level"))



from nose.tools import *
from tests.tablespaces.tablespace_utils import node2


# Node 1 is paused by the Makefile, wait for pgaf to acknowledge
def test_001_old_primary_goes_down():
    assert node2.wait_until_state(target_state="wait_primary")

    node2.run_sql_query(
        "CREATE TABLESPACE extended_c LOCATION '/extra_volumes/extended_c';"
    )
    node2.run_sql_query("CREATE TABLE t4(i int) TABLESPACE extended_c;")
    node2.run_sql_query("INSERT INTO t4 VALUES (10), (11)")
    node2.run_sql_query("INSERT INTO t2 VALUES (12)")
    node2.run_sql_query("INSERT INTO t3 VALUES (13)")



import tests.pgautofailover_utils as pgautofailover
import tests.ssl_cert_utils as cert
import subprocess
import os
import time

cluster = None
monitor = None
node1 = None
node2 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()

    # remove client side setup for certificates too
    client_top_directory = os.path.join(os.getenv("HOME"), ".postgresql")

    p = subprocess.Popen(
        [
            "sudo",
            "-E",
            "-u",
            os.getenv("USER"),
            "env",
            "PATH=" + os.getenv("PATH"),
            "rm",
            "-rf",
            client_top_directory,
        ]
    )
    assert p.wait() == 0

    # also remove certificates we created for the servers
    p = subprocess.run(
        [
            "sudo",
            "-E",
            "-u",
            os.getenv("USER"),
            "env",
            "PATH=" + os.getenv("PATH"),
            "rm",
            "-rf",
            "/tmp/certs",
        ]
    )
    assert p.returncode == 0


def test_000_create_monitor():
    global monitor
    monitor = cluster.create_monitor("/tmp/enable/monitor")
    monitor.run()
    monitor.wait_until_pg_is_running()

    monitor.check_ssl("off", "prefer")


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/enable/node1")
    node1.create()
    node1.run()
    assert node1.wait_until_state(target_state="single")

    node1.wait_until_pg_is_running()
    node1.check_ssl("off", "prefer", primary=True)


def test_002_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_init_secondary():
    global node2
    node2 = cluster.create_datanode("/tmp/enable/node2")
    node2.create()
    node2.run()

    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    node2.check_ssl("off", "prefer")


def test_004_maintenance():
    print()
    print("Enabling maintenance on node2")
    node2.enable_maintenance()
    assert node2.wait_until_state(target_state="maintenance")


def test_005_enable_ssl_monitor():
    monitor.enable_ssl(sslSelfSigned=True, sslMode="require")
    monitor.sleep(2)  # we signaled, wait some time

    monitor.check_ssl("on", "require")


def test_006_enable_ssl_primary():
    # we stop pg_autoctl to make it easier for the test to be reliable
    # without too much delay/sleep hacking; when doing the `pg_autoctl
    # enable ssl` online we need to make sure the signal made it to the
    # running process and then was acted upon
    node1.stop_pg_autoctl()
    node1.enable_ssl(sslSelfSigned=True, sslMode="require")
    node1.run()

    node1.wait_until_pg_is_running()
    node1.check_ssl("on", "require", primary=True)


def test_007_enable_ssl_secondary():
    node2.stop_pg_autoctl()
    node2.enable_ssl(sslSelfSigned=True, sslMode="require")
    node2.run()

    node2.wait_until_pg_is_running()
    node2.check_ssl("on", "require")


def test_008_disable_maintenance():
    print("Disabling maintenance on node2")
    node2.disable_maintenance()
    assert node2.wait_until_pg_is_running()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


# upgrade to verify full
def test_009_enable_maintenance():
    print()
    print("Enabling maintenance on node2")
    node2.enable_maintenance()

    assert node2.wait_until_state(target_state="maintenance")


def test_010_enable_ssl_verify_ca_monitor():
    client_top_directory = os.path.join(os.getenv("HOME"), ".postgresql")

    print()
    print("Creating cluster root certificate")
    cluster.create_root_cert(
        client_top_directory, basename="root", CN="/CN=root.pgautofailover.ca"
    )

    p = subprocess.run(
        [
            "ls",
            "-ld",
            client_top_directory,
            cluster.cert.crt,
            cluster.cert.csr,
            cluster.cert.key,
        ],
        text=True,
        capture_output=True,
    )
    print("%s" % p.stdout)

    # now create and sign the CLIENT certificate
    print("Creating cluster client certificate")
    clientCert = cert.SSLCert(
        client_top_directory, basename="postgresql", CN="/CN=autoctl_node"
    )
    clientCert.create_signed_certificate(cluster.cert)

    p = subprocess.run(
        [
            "ls",
            "-ld",
            client_top_directory,
            clientCert.crt,
            clientCert.csr,
            clientCert.key,
        ],
        text=True,
        capture_output=True,
    )
    print("%s" % p.stdout)

    # the root user also needs the certificates, tests are connecting with it
    root_top_directory = "/root/.postgresql"
    p = subprocess.run(
        ["sudo", "install", "-d", "-m", "740", root_top_directory]
    )
    assert p.returncode == 0

    p = subprocess.run(
        [
            "sudo",
            "cp",
            clientCert.crt,
            clientCert.csr,
            clientCert.key,
            root_top_directory,
        ]
    )
    assert p.returncode == 0

    p = subprocess.run(
        ["ls", "-l", "/root/.postgresql"], text=True, capture_output=True
    )
    print("%s" % p.stdout)

    # now create and sign the SERVER certificate for the monitor
    print("Creating monitor server certificate")
    monitorCert = cert.SSLCert(
        "/tmp/certs/monitor", "server", "/CN=monitor.pgautofailover.ca"
    )
    monitorCert.create_signed_certificate(cluster.cert)

    p = subprocess.run(
        [
            "ls",
            "-ld",
            client_top_directory,
            cluster.cert.crt,
            cluster.cert.csr,
            cluster.cert.key,
            clientCert.crt,
            clientCert.csr,
            clientCert.key,
            monitorCert.crt,
            monitorCert.csr,
            monitorCert.key,
        ],
        text=True,
        capture_output=True,
    )
    print("%s" % p.stdout)

    monitor.enable_ssl(
        sslCAFile=cluster.cert.crt,
        sslServerKey=monitorCert.key,
        sslServerCert=monitorCert.crt,
        sslMode="verify-ca",
    )

    monitor.sleep(2)  # we signaled, wait some time

    monitor.check_ssl("on", "verify-ca")


def test_011_enable_ssl_verify_ca_primary():
    node1Cert = cert.SSLCert(
        "/tmp/certs/node1", "server", "/CN=node1.pgautofailover.ca"
    )
    node1Cert.create_signed_certificate(cluster.cert)

    node1.stop_pg_autoctl()
    node1.enable_ssl(
        sslCAFile=cluster.cert.crt,
        sslServerKey=node1Cert.key,
        sslServerCert=node1Cert.crt,
        sslMode="verify-ca",
    )
    node1.run()
    node1.wait_until_pg_is_running()
    node1.check_ssl("on", "verify-ca", primary=True)


def test_012_enable_ssl_verify_ca_secondary():
    node2Cert = cert.SSLCert(
        "/tmp/certs/node2", "server", "/CN=node2.pgautofailover.ca"
    )
    node2Cert.create_signed_certificate(cluster.cert)

    node2.stop_pg_autoctl()
    node2.enable_ssl(
        sslCAFile=cluster.cert.crt,
        sslServerKey=node2Cert.key,
        sslServerCert=node2Cert.crt,
        sslMode="verify-ca",
    )
    node2.run()
    node2.wait_until_pg_is_running()
    node2.check_ssl("on", "verify-ca")


def test_013_disable_maintenance():
    print("Disabling maintenance on node2")
    node2.disable_maintenance()
    assert node2.wait_until_pg_is_running()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


def test_014_enable_ssl_require_primary():
    node1Cert = cert.SSLCert(
        "/tmp/certs/node1", "server", "/CN=node1.pgautofailover.ca"
    )
    node1Cert.create_signed_certificate(cluster.cert)

    node1.enable_ssl(
        sslServerKey=node1Cert.key,
        sslServerCert=node1Cert.crt,
        sslMode="require",
    )

    node1.pg_autoctl.sighup()
    time.sleep(6)

    # to avoid flackyness here, we allow a second run/timeout of waiting
    if not node1.wait_until_pg_is_running():
        assert node1.wait_until_pg_is_running()

    node1.check_ssl("on", "require", primary=True)



import tests.pgautofailover_utils as pgautofailover
import os.path
import subprocess

cluster = None
monitor = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    global monitor

    print()
    monitor_path = cluster.pg_createcluster("monitor", port=6000)

    postgres_conf_path = os.path.join(monitor_path, "postgresql.conf")
    # verify postgresql.conf is not in data directory
    assert not os.path.exists(postgres_conf_path)

    monitor = cluster.create_monitor(monitor_path, port=6000)
    monitor.create(level="-vv")
    monitor.run(port=6000)
    monitor.wait_until_pg_is_running()

    # verify postgresql.conf is in data directory now
    assert os.path.exists(postgres_conf_path)

    pgversion = os.getenv("PGVERSION")

    p = subprocess.run(
        [
            "ls",
            "-ld",
            monitor_path,
            "/var/lib/postgresql/%s" % pgversion,
            "/etc/postgresql/%s" % pgversion,
            "/etc/postgresql/%s/monitor" % pgversion,
            "/etc/postgresql/%s/monitor/postgresql.conf" % pgversion,
            "/etc/postgresql/%s/monitor/pg_hba.conf" % pgversion,
            "/etc/postgresql/%s/monitor/pg_ident.conf" % pgversion,
        ],
        text=True,
        capture_output=True,
    )
    print("%s" % p.stdout)


def test_001_custom_single():
    global node1

    print()
    node1_path = cluster.pg_createcluster("debian_node1", port=6001)

    postgres_conf_path = os.path.join(node1_path, "postgresql.conf")
    # verify postgresql.conf is not in data directory
    assert not os.path.exists(postgres_conf_path)

    node1 = cluster.create_datanode(node1_path, port=6001, listen_flag=True)
    node1.create(level="-vv")

    # verify postgresql.conf is in data directory now
    assert os.path.exists(postgres_conf_path)

    pgversion = os.getenv("PGVERSION")

    p = subprocess.run(
        [
            "ls",
            "-ld",
            node1_path,
            "/var/lib/postgresql/%s" % pgversion,
            "/etc/postgresql/%s" % pgversion,
            "/etc/postgresql/%s/debian_node1" % pgversion,
            "/etc/postgresql/%s/debian_node1/postgresql.conf" % pgversion,
            "/etc/postgresql/%s/debian_node1/pg_hba.conf" % pgversion,
            "/etc/postgresql/%s/debian_node1/pg_ident.conf" % pgversion,
        ],
        text=True,
        capture_output=True,
    )
    print("%s" % p.stdout)

    monitor.print_state()


def test_002_chmod_debian_data_directory():
    # debian installs the following ownership and permissions:
    #
    # drwxr-xr-x  5 postgres postgres ... /var/lib/postgresql/11
    # drwx------ 20 docker   postgres ... /var/lib/postgresql/11/monitor
    # drwx------ 20 docker   postgres ... /var/lib/postgresql/11/debian_node1
    #
    # we need to give the postgres group the w on the top-level directory
    pgversion = os.getenv("PGVERSION")

    p = subprocess.Popen(
        ["chmod", "go+w", "/var/lib/postgresql/%s" % pgversion]
    )
    assert p.wait() == 0



import tests.pgautofailover_utils as pgautofailover
from nose.tools import *

import subprocess
import shutil
import os
import os.path

cluster = None
node1 = None
node2 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    monitor = cluster.create_monitor("/tmp/check/monitor")
    monitor.run()
    monitor.wait_until_pg_is_running()


def test_001_add_hba_entry():
    with open(os.path.join("/tmp/check/monitor", "pg_hba.conf"), "a") as hba:
        hba.write("host all all %s trust\n" % cluster.networkSubnet)

    # print()
    # with open(os.path.join("/tmp/check/monitor", "pg_hba.conf"), "r") as hba:
    #     lines = hba.readlines()
    #     for line in lines[-10:]:
    #         print("%s" % line[:-1])

    cluster.monitor.reload_postgres()


def test_002_make_installcheck():
    # support both the local Dockerfile and also Travis build environments
    if "TRAVIS_BUILD_DIR" in os.environ:
        topdir = os.environ["TRAVIS_BUILD_DIR"]
    else:
        topdir = "/usr/src/pg_auto_failover"

    p = subprocess.Popen(
        [
            "sudo",
            shutil.which("chmod"),
            "-R",
            "go+w",
            os.path.join(topdir, "src/monitor"),
        ]
    )
    assert p.wait() == 0

    p = subprocess.Popen(
        [
            "sudo",
            "-E",
            "-u",
            os.getenv("USER"),
            "env",
            "PATH=" + os.getenv("PATH"),
            "PGHOST=" + str(cluster.monitor.vnode.address),
            "make",
            "-C",
            os.path.join(topdir, "src/monitor"),
            "installcheck",
        ]
    )

    if p.wait() != 0:
        diff = os.path.join(topdir, "src/monitor/regression.diffs")
        with open(diff, "r") as d:
            print("%s" % d.read())

        raise Exception("make installcheck failed")



from nose.tools import *
from tests.tablespaces.tablespace_utils import node1


# Node 1 is spun up by Makefile, wait until it's up
def test_000_init_monitor_and_primary():
    # Wait a bit longer than normal for start up
    assert node1.wait_until_state(
        target_state="single", timeout=90, sleep_time=3
    )


def test_002_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_create_tablespace():
    node1.run_sql_query(
        "CREATE TABLESPACE extended_a LOCATION '/extra_volumes/extended_a';"
    )
    node1.run_sql_query("CREATE TABLE t2(i int) TABLESPACE extended_a;")
    node1.run_sql_query("INSERT INTO t2 VALUES (3), (4);")



import tests.pgautofailover_utils as pgautofailover
from nose.tools import *

cluster = None
monitor = None
coordinator1a = None
coordinator1b = None
worker1a = None
worker1b = None
worker2a = None
worker2b = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    if coordinator1b is not None:
        coordinator1b.run_sql_query("select public.wait_until_metadata_sync()")
        coordinator1b.run_sql_query("DROP TABLE t1")
    cluster.destroy()


def test_000_create_monitor():
    global monitor
    monitor = cluster.create_monitor("/tmp/nonha/monitor")
    monitor.run()


def test_001_create_formation():
    global monitor

    monitor.create_formation("non-ha", kind="citus", secondary=False)


def test_002_init_coordinator():
    global coordinator1a

    coordinator1a = cluster.create_datanode(
        "/tmp/citus/nonha/coordinator1a",
        role=pgautofailover.Role.Coordinator,
        formation="non-ha",
    )
    coordinator1a.create()
    coordinator1a.run()

    print()  # make the debug output more readable
    assert coordinator1a.wait_until_state(target_state="single")

    # we need to expose some Citus testing internals
    assert coordinator1a.wait_until_pg_is_running()
    coordinator1a.create_wait_until_metadata_sync()


def test_003_init_workers():
    global worker1a
    global worker2a

    worker1a = cluster.create_datanode(
        "/tmp/citus/nonha/worker1a",
        role=pgautofailover.Role.Worker,
        group=1,
        formation="non-ha",
    )
    worker1a.create()
    worker1a.run()

    worker2a = cluster.create_datanode(
        "/tmp/citus/nonha/worker2a",
        role=pgautofailover.Role.Worker,
        group=2,
        formation="non-ha",
    )
    worker2a.create()
    worker2a.run()

    # wait here till all workers are stable and in the desired state
    # by not waiting on all workers separately this should save a bit of time
    print()  # make the debug output more readable
    assert worker1a.wait_until_state(target_state="single")
    assert worker2a.wait_until_state(target_state="single")


def test_004_create_distributed_table():
    assert coordinator1a.wait_until_pg_is_running()
    coordinator1a.run_sql_query("CREATE TABLE t1 (a int)")
    coordinator1a.run_sql_query("SELECT create_distributed_table('t1', 'a')")
    coordinator1a.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_005_enable_secondary():
    global monitor
    monitor.enable(pgautofailover.Feature.Secondary, formation="non-ha")


def test_006_add_secondaries():
    global coordinator1b
    global worker1b
    global worker2b

    coordinator1b = cluster.create_datanode(
        "/tmp/citus/nonha/coordinator1b",
        role=pgautofailover.Role.Coordinator,
        formation="non-ha",
    )
    coordinator1b.create()
    coordinator1b.run()

    worker1b = cluster.create_datanode(
        "/tmp/citus/nonha/worker1b",
        role=pgautofailover.Role.Worker,
        group=1,
        formation="non-ha",
    )
    worker1b.create()
    worker1b.run()

    worker2b = cluster.create_datanode(
        "/tmp/citus/nonha/worker2b",
        role=pgautofailover.Role.Worker,
        group=2,
        formation="non-ha",
    )
    worker2b.create()
    worker2b.run()
    # wait here till all workers are stable and in the desired state
    # by not waiting on all workers separately this should save a bit of time
    print()  # make the debug output more readable
    assert coordinator1b.wait_until_state(target_state="secondary")
    assert worker1b.wait_until_state(target_state="secondary")
    assert worker2b.wait_until_state(target_state="secondary")


@raises(Exception)
def test_007_fail_when_disabling_with_secondaries():
    global monitor
    monitor.disable(pgautofailover.Feature.Secondary, formation="non-ha")


def test_008_shutdown_primaries():
    print()  # make the debug output more readable
    coordinator1a.fail()
    assert coordinator1b.wait_until_state(target_state="wait_primary")

    worker1a.fail()
    assert worker1b.wait_until_state(target_state="wait_primary")

    worker2a.fail()
    assert worker2b.wait_until_state(target_state="wait_primary")


def test_009_remove_failed_nodes():
    coordinator1a.drop()
    worker1a.drop()
    worker2a.drop()

    print()  # make the debug output more readable
    assert coordinator1b.wait_until_state(target_state="single")
    assert worker1b.wait_until_state(target_state="single")
    assert worker2b.wait_until_state(target_state="single")


def test_010_disable_secondaries():
    global monitor
    monitor.disable(pgautofailover.Feature.Secondary, formation="non-ha")



#
# SSL Certificate creation in the test environment
#
import subprocess
import os, os.path, time, shutil


class SSLCert:
    """
    Calls openssl to generate SSL certificates and sign them.
    """

    def __init__(self, directory, basename, CN):
        self.directory = directory
        self.basename = basename
        self.CN = CN

        self.csr = None
        self.key = None
        self.crt = None

        self.rootKey = None
        self.rootCert = None
        self.crl = None

        self.sudo_mkdir_p()

    def sudo_mkdir_p(self):
        if os.path.isdir(self.directory):
            return True

        p = subprocess.Popen(
            [
                "sudo",
                "-E",
                "-u",
                os.getenv("USER"),
                "env",
                "PATH=" + os.getenv("PATH"),
                "mkdir",
                "-p",
                self.directory,
            ]
        )
        assert p.wait() == 0

    def create_root_cert(self):
        # avoid bugs where we overwrite certificates in a given directory
        assert self.csr is None
        assert self.crt is None
        assert self.key is None

        self.csr = os.path.join(self.directory, "%s.csr" % self.basename)
        self.key = os.path.join(self.directory, "%s.key" % self.basename)
        self.crt = os.path.join(self.directory, "%s.crt" % self.basename)

        # first create a certificate signing request (CSR) and a public/private
        # key file
        print()
        p = subprocess.Popen(
            [
                "sudo",
                "-E",
                "-u",
                os.getenv("USER"),
                "env",
                "PATH=" + os.getenv("PATH"),
                "openssl",
                "req",
                "-new",
                "-nodes",
                "-text",
                "-out",
                self.csr,
                "-keyout",
                self.key,
                "-subj",
                self.CN,
            ]
        )
        assert p.wait() == 0

        p = subprocess.Popen(["chmod", "og-rwx", self.key])
        assert p.wait() == 0

        # Then, sign the request with the key to create a root certificate
        # authority
        p = subprocess.Popen(
            [
                "sudo",
                "-E",
                "-u",
                os.getenv("USER"),
                "env",
                "PATH=" + os.getenv("PATH"),
                "openssl",
                "x509",
                "-req",
                "-in",
                self.csr,
                "-text",
                "-days",
                "3650",
                "-extfile",
                "/etc/ssl/openssl.cnf",
                "-extensions",
                "v3_ca",
                "-signkey",
                self.key,
                "-out",
                self.crt,
            ]
        )
        assert p.wait() == 0

    def create_signed_certificate(self, rootSSLCert):
        # avoid bugs where we overwrite certificates in a given directory
        assert self.csr is None
        assert self.crt is None
        assert self.key is None

        self.rootKey = rootSSLCert.key
        self.rootCert = rootSSLCert.crt
        self.crl = rootSSLCert.crl

        self.crt = os.path.join(self.directory, "%s.crt" % self.basename)
        self.csr = os.path.join(self.directory, "%s.csr" % self.basename)
        self.key = os.path.join(self.directory, "%s.key" % self.basename)

        p = subprocess.Popen(
            [
                "sudo",
                "-E",
                "-u",
                os.getenv("USER"),
                "env",
                "PATH=" + os.getenv("PATH"),
                "openssl",
                "req",
                "-new",
                "-nodes",
                "-text",
                "-out",
                self.csr,
                "-keyout",
                self.key,
                "-subj",
                self.CN,
            ]
        )
        assert p.wait() == 0

        p = subprocess.Popen(["chmod", "og-rwx", self.key])
        assert p.wait() == 0

        p = subprocess.Popen(
            [
                "sudo",
                "-E",
                "-u",
                os.getenv("USER"),
                "env",
                "PATH=" + os.getenv("PATH"),
                "openssl",
                "x509",
                "-req",
                "-in",
                self.csr,
                "-text",
                "-days",
                "365",
                "-CA",
                self.rootCert,
                "-CAkey",
                self.rootKey,
                "-CAcreateserial",
                "-out",
                self.crt,
            ]
        )
        assert p.wait() == 0

        print("openssl verify -CAfile %s %s" % (self.rootCert, self.crt))
        p = subprocess.Popen(
            [
                "sudo",
                "-E",
                "-u",
                os.getenv("USER"),
                "env",
                "PATH=" + os.getenv("PATH"),
                "openssl",
                "verify",
                "-show_chain",
                "-CAfile",
                self.rootCert,
                self.crt,
            ]
        )
        assert p.wait() == 0



import tests.pgautofailover_utils as pgautofailover
from nose.tools import *

import os
import re

cluster = None
node1 = None
node2 = None
replication_password = "streaming_password"
monitor_password = "monitor_password"


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    monitor = cluster.create_monitor("/tmp/auth/monitor", authMethod="md5")
    monitor.run()
    monitor.wait_until_pg_is_running()
    monitor.set_user_password("autoctl_node", "autoctl_node_password")

    monitor.create_formation("auth", kind="pgsql", secondary=True)


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode(
        "/tmp/auth/node1", authMethod="md5", formation="auth"
    )
    node1.create()
    node1.config_set("replication.password", replication_password)
    node1.run()

    node1.wait_until_pg_is_running()
    node1.set_user_password("pgautofailover_monitor", monitor_password)
    node1.set_user_password("pgautofailover_replicator", replication_password)

    assert node1.wait_until_state(target_state="single")


def test_002_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_init_secondary():
    global node2
    node2 = cluster.create_datanode(
        "/tmp/auth/node2", authMethod="md5", formation="auth"
    )

    os.putenv("PGPASSWORD", replication_password)
    node2.create()
    node2.config_set("replication.password", replication_password)

    node2.run()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    eq_(
        node1.get_synchronous_standby_names_local(),
        "ANY 1 (pgautofailover_standby_2)",
    )

    # Prevent regression from bug fixed in PR #839
    logs = node2.logs("STDERR")
    match = re.search("Failed to CHECKPOINT before restart", logs)
    if match:
        print("Found CHECKPOINT in logs: ", match[0])
    assert not match
    node2.run()


def test_004_failover():
    print()
    print("Calling pgautofailover.failover() on the monitor")
    cluster.monitor.failover(formation="auth")
    assert node2.wait_until_state(target_state="primary")
    eq_(
        node2.get_synchronous_standby_names_local(),
        "ANY 1 (pgautofailover_standby_1)",
    )

    assert node1.wait_until_state(target_state="secondary")


def test_005_logging_of_passwords():
    logs = node2.logs()
    assert monitor_password not in logs
    assert "password=****" in logs

    # We are still logging passwords when the pguri is incomplete and when
    # printing settings, so assert that it's not there in other cases:
    assert not re.search(
        "^(?!primary_conninfo|Failed to find).*%s.*$" % replication_password,
        logs,
    )



from pyroute2 import netns, NDB, netlink, NSPopen
from contextlib import contextmanager
import ipaddress
import subprocess
import os
import os.path

"""
TODO: Add an introduction to network namespaces, veth interfaces, and bridges,
and explain why we use them here.
"""

BRIDGE_NF_CALL_IPTABLES = "/proc/sys/net/bridge/bridge-nf-call-iptables"
COMMAND_TIMEOUT = 60


@contextmanager
def managed_nspopen(*args, **kwds):
    proc = NSPopen(*args, **kwds)
    try:
        yield proc
    finally:
        if proc.poll() is None:
            # send SIGKILL to the process and wait for it to die if it's still
            # running
            proc.kill()
            # If it's not dead after 2 seconds we throw an error
            proc.communicate(timeout=2)

        # release proxy process resourecs
        proc.release()


class VirtualLAN:
    """
    Helper class to create a network of virtual nodes to simulate a virtual network.
    IP addresses are assigned automatically to the nodes from a private IP range.
    IP address of a virtual node can be accessed using the node.address field.

    Internally, this is a network of Linux network namespaces connected by a
    bridge.
    TODO: explain more details and add an example.
    """

    def __init__(self, namePrefix, subnet):
        ipnet = ipaddress.ip_network(subnet)
        self.availableHosts = ipnet.hosts()
        self.prefixLen = ipnet.prefixlen
        self.namePrefix = namePrefix
        self.nodes = []
        # create the bridge
        self.bridgeName = "%s-br" % (namePrefix,)
        self.bridgeAddress = next(self.availableHosts)
        self._add_bridge(self.bridgeName, self.bridgeAddress, self.prefixLen)
        # Don't pass bridged IPv4 traffic to iptables' chains, so namespaces
        # can communicate irrespective of the host machines iptables. This is
        # needed in some docker instances (e.g. travis), where traffic was
        # filtered at bridge level. See
        # https://www.kernel.org/doc/Documentation/networking/ip-sysctl.txt
        try:
            with open(BRIDGE_NF_CALL_IPTABLES, "r") as f:
                self.saved_bridge_nf_call_iptables = f.read()
            with open(BRIDGE_NF_CALL_IPTABLES, "w") as f:
                f.write("0\n")
        except FileNotFoundError:
            # In some environments this variable doesn't exist, we are ok with
            # no changes in this case.
            self.saved_bridge_nf_call_iptables = None

    def create_node(self):
        """
        Creates a VirtualNode which can access/be accessed from other nodes in
        the virtual network.
        """
        namespace = "%s-%s" % (self.namePrefix, len(self.nodes))
        address = next(self.availableHosts)
        node = VirtualNode(namespace, address, self.prefixLen)
        self._add_interface_to_bridge(self.bridgeName, node.vethPeer)
        self.nodes.append(node)
        return node

    def destroy(self):
        """
        Destroys the objects created for the virtual network.
        """
        for node in self.nodes:
            node.destroy()
        _remove_interface_if_exists(self.bridgeName)
        if self.saved_bridge_nf_call_iptables is not None:
            with open(BRIDGE_NF_CALL_IPTABLES, "w") as f:
                f.write(self.saved_bridge_nf_call_iptables)

    def _add_bridge(self, name, address, prefixLen):
        """
        Creates a bridge with the given name, address, and netmask perfix length.
        """
        _remove_interface_if_exists(name)

        with NDB() as ndb:
            (
                ndb.interfaces.create(ifname=name, kind="bridge", state="up")
                .add_ip("%s/%s" % (address, prefixLen))
                .commit()
            )

    def _add_interface_to_bridge(self, bridge, interface):
        """
        Adds the given interface to the bridge. In our usecase, this interface
        is usually the peer end of a veth pair with the other end inside a
        network namespace, in which case after calling this function the namespace
        will be able to communicate with the other nodes in the virtual network.
        """
        with NDB() as ndb:
            ndb.interfaces[bridge].add_port(interface).commit()
            ndb.interfaces[interface].set(state="up").commit()


class VirtualNode:
    """
    A virtual node inside a virtual network.

    Internally, this corresponds to a Linux network namespace.
    """

    def __init__(self, namespace, address, prefixLen):
        self.namespace = namespace
        self.address = address
        self.prefixLen = prefixLen
        self.vethPeer = namespace + "p"
        self._add_namespace(namespace, address, prefixLen)

    def destroy(self):
        """
        Removes all objects created for the virtual node.
        """
        _remove_interface_if_exists(self.vethPeer)
        try:
            netns.remove(self.namespace)
        except:
            # Namespace doesn't exist. Return silently.
            pass

    def run(self, command, user=os.getenv("USER")):
        """
        Executes a command under the given user from this virtual node. Returns
        a context manager that returns NSOpen object to control the process.
        NSOpen has the same API as subprocess.POpen.
        """
        sudo_command = [
            "sudo",
            "-E",
            "-u",
            user,
            "env",
            "PATH=" + os.getenv("PATH"),
        ] + command
        return managed_nspopen(
            self.namespace,
            sudo_command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            start_new_session=True,
        )

    def run_unmanaged(self, command, user=os.getenv("USER")):
        """
        Executes a command under the given user from this virtual node. Returns
        an NSPopen object to control the process. NSOpen has the same API as
        subprocess.Popen. This NSPopen object needs to be manually release. In
        general you should prefer using run, where this is done automatically
        by the context manager.
        """
        sudo_command = [
            "sudo",
            "-E",
            "-u",
            user,
            "env",
            "PATH=" + os.getenv("PATH"),
        ] + command
        return NSPopen(
            self.namespace,
            sudo_command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            start_new_session=True,
        )

    def run_and_wait(self, command, name, timeout=COMMAND_TIMEOUT):
        """
        Waits for command to exit successfully. If it exits with error or it timeouts,
        raises an exception with stdout and stderr streams of the process.
        """
        with self.run(command) as proc:
            try:
                out, err = proc.communicate(timeout=timeout)
                if proc.returncode > 0:
                    raise Exception(
                        "%s failed, out: %s\n, err: %s" % (name, out, err)
                    )
                return out, err
            except subprocess.TimeoutExpired:
                proc.kill()
                out, err = proc.communicate()
                raise Exception(
                    "%s timed out after %d seconds. out: %s\n, err: %s"
                    % (name, timeout, out, err)
                )

    def _add_namespace(self, name, address, netmaskLength):
        """
        Creates a namespace with the given name, and creates a veth interface
        with one endpoint inside the namespace which has the given address and
        netmask length. The peer end of veth interface can be used to connect the
        namespace to a bridge.
        """

        self._remove_namespace_if_exists(name)
        netns.create(name)

        veth_name = "veth0"

        _remove_interface_if_exists(self.vethPeer)

        with NDB() as ndb:
            #
            # Add netns to the NDB sources
            #
            # ndb.interfaces["lo"] is a short form of
            # ndb.interfaces[{"target": "localhost", "ifname": "lo"}]
            #
            # To address interfaces/addresses/routes wthin a netns, use
            # ndb.interfaces[{"target": netns_name, "ifname": "lo"}]
            ndb.sources.add(netns=name)
            #
            # Create veth
            (
                ndb.interfaces.create(
                    ifname=veth_name,
                    kind="veth",
                    peer=self.vethPeer,
                    state="up",
                )
                .commit()
                .set(net_ns_fd=name)
                .commit()
            )
            #
            # .interfaces.wait() returns an interface object when
            # it becomes available on the specified source
            (
                ndb.interfaces.wait(target=name, ifname=veth_name)
                .set(state="up")
                .add_ip("%s/%s" % (address, netmaskLength))
                .commit()
            )
            #
            (
                ndb.interfaces[{"target": name, "ifname": "lo"}]
                .set(state="up")
                .commit()
            )

    def _remove_namespace_if_exists(self, name):
        """
        If the given namespace exists, removes it. Otherwise just returns
        silently.
        """
        try:
            netns.remove(name)
        except Exception:
            # Namespace doesn't exist. Return silently.
            pass

    def ifdown(self):
        """
        Bring the network interface down for this node
        """
        with NDB() as ndb:
            # bring it down and wait until success
            ndb.interfaces[self.vethPeer].set(state="down").commit()

    def ifup(self):
        """
        Bring the network interface up for this node
        """
        with NDB() as ndb:
            # bring it up and wait until success
            ndb.interfaces[self.vethPeer].set(state="up").commit()


def _remove_interface_if_exists(name):
    """
    If the given interface exists, brings it down and removes it. Otherwise
    just returns silently. A bridge is also an interface, so this can be
    used for removing bridges too.
    """
    with NDB() as ndb:
        if name in ndb.interfaces:
            try:
                ndb.interfaces[name].remove().commit()
            except netlink.exceptions.NetlinkError:
                pass



import tests.pgautofailover_utils as pgautofailover
import time

from nose.tools import eq_

cluster = None
monitor = None
node1 = None
node2 = None

newmonitor = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    global monitor
    monitor = cluster.create_monitor("/tmp/replace/monitor")
    monitor.run()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/replace/node1")
    node1.create(run=True)
    assert node1.wait_until_state(target_state="single")


def test_002_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_init_secondary():
    global node2
    node2 = cluster.create_datanode("/tmp/replace/node2")
    node2.create(run=True)
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


def test_004_read_from_secondary():
    results = node2.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,)]


def test_005_drop_monitor():
    monitor.destroy()
    cluster.monitor = None


def test_006a_disable_monitor_node2():
    node2.disable_monitor()


def test_006b_disable_monitor_node1():
    node1.disable_monitor()


def test_006c_write_to_primary():
    node1.run_sql_query("INSERT INTO t1 VALUES (3)")


def test_007_read_from_secondary():
    results = node2.run_sql_query("SELECT * FROM t1 ORDER BY a")
    assert results == [(1,), (2,), (3,)]


def test_008_create_new_monitor():
    global newmonitor
    newmonitor = cluster.create_monitor("/tmp/replace/newmonitor")
    newmonitor.run()


def test_009a_enable_monitor_node1():
    node1.enable_monitor(newmonitor)
    assert node1.wait_until_state(target_state="single")


def test_009b_enable_monitor_node2():
    node2.enable_monitor(newmonitor)
    assert node2.wait_until_state(target_state="catchingup")
    assert node1.wait_until_state(target_state="wait_primary")


def test_010_wait_until_state():
    assert node1.wait_until_state(target_state="primary")
    assert node2.wait_until_state(target_state="secondary")


def test_011_failover():
    print()
    print("Calling pgautofailover.failover() on the monitor")
    newmonitor.failover()

    assert node2.wait_until_state(target_state="primary")
    eq_(
        node2.get_synchronous_standby_names_local(),
        "ANY 1 (pgautofailover_standby_1)",
    )

    assert node1.wait_until_state(target_state="secondary")


def test_012_read_from_secondary():
    results = node1.run_sql_query("SELECT * FROM t1 ORDER BY a")
    assert results == [(1,), (2,), (3,)]



from nose.tools import *
from tests.tablespaces.tablespace_utils import node1
from tests.tablespaces.tablespace_utils import node2


# Failover is performed by the makefile Makefile, wait for it to complete
def test_001_promote_the_original_primary_successfully():
    assert node1.wait_until_state(target_state="primary")
    assert node2.wait_until_state(target_state="secondary")



import tests.pgautofailover_utils as pgautofailover
import tests.ssl_cert_utils as cert
from nose.tools import *

import subprocess
import os, os.path, time, shutil

cluster = None
node1 = None
node2 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()

    client_top_directory = os.path.join(os.getenv("HOME"), ".postgresql")

    cluster.create_root_cert(
        client_top_directory, basename="root", CN="/CN=root.pgautofailover.ca"
    )


def teardown_module():
    cluster.destroy()

    # remove client side setup for certificates too
    client_top_directory = os.path.join(os.getenv("HOME"), ".postgresql")

    p = subprocess.Popen(
        [
            "sudo",
            "-E",
            "-u",
            os.getenv("USER"),
            "env",
            "PATH=" + os.getenv("PATH"),
            "rm",
            "-rf",
            client_top_directory,
        ]
    )
    assert p.wait() == 0

    # also remove certificates we created for the servers
    p = subprocess.run(
        [
            "sudo",
            "-E",
            "-u",
            os.getenv("USER"),
            "env",
            "PATH=" + os.getenv("PATH"),
            "rm",
            "-rf",
            "/tmp/certs",
        ]
    )
    assert p.returncode == 0


def test_000_create_monitor():
    # create SSL certs and keys for this server
    #
    # https://www.postgresql.org/docs/11/ssl-tcp.html
    #
    # server.crt and server.key should be stored on the server, and root.crt
    # should be stored on the client so the client can verify that the
    # server's leaf certificate was signed by its trusted root certificate.
    # root.key should be stored offline for use in creating future
    # certificates.
    #
    # https://www.postgresql.org/docs/current/libpq-ssl.html
    #
    # If the server attempts to verify the identity of the client by
    # requesting the client's leaf certificate, libpq will send the
    # certificates stored in file ~/.postgresql/postgresql.crt in the user's
    # home directory
    # Now, create a server certificate signed by the new root certificate
    # authority
    client_top_directory = os.path.join(os.getenv("HOME"), ".postgresql")

    # now create and sign the CLIENT certificate
    clientCert = cert.SSLCert(
        client_top_directory, basename="postgresql", CN="/CN=autoctl_node"
    )
    clientCert.create_signed_certificate(cluster.cert)

    # now create and sign the SERVER certificate for the monitor
    serverCert = cert.SSLCert(
        "/tmp/certs/monitor", "server", "/CN=monitor.pgautofailover.ca"
    )
    serverCert.create_signed_certificate(cluster.cert)

    # the root user also needs the certificates, tests are connecting with it
    root_top_directory = "/root/.postgresql"
    p = subprocess.run(
        ["sudo", "install", "-d", "-m", "740", root_top_directory]
    )
    assert p.returncode == 0

    p = subprocess.run(
        [
            "sudo",
            "cp",
            clientCert.crt,
            clientCert.csr,
            clientCert.key,
            root_top_directory,
        ]
    )
    assert p.returncode == 0

    p = subprocess.run(
        [
            "ls",
            "-ld",
            client_top_directory,
            root_top_directory,
            os.path.join(root_top_directory, "postgresql.crt"),
            os.path.join(root_top_directory, "postgresql.csr"),
            os.path.join(root_top_directory, "postgresql.key"),
            cluster.cert.crt,
            cluster.cert.csr,
            cluster.cert.key,
            clientCert.crt,
            clientCert.csr,
            clientCert.key,
            serverCert.crt,
            serverCert.csr,
            serverCert.key,
        ],
        text=True,
        capture_output=True,
    )
    print("%s" % p.stdout)

    #
    # Now create the monitor Postgres instance with the certificates
    #
    monitor = cluster.create_monitor(
        "/tmp/cert/monitor",
        authMethod="skip",
        sslMode="verify-ca",
        sslCAFile=cluster.cert.crt,
        sslServerKey=serverCert.key,
        sslServerCert=serverCert.crt,
    )
    monitor.run()
    monitor.wait_until_pg_is_running()

    with open(os.path.join("/tmp/cert/monitor", "pg_hba.conf"), "a") as hba:
        hba.write("hostssl all all %s cert\n" % cluster.networkSubnet)

    monitor.reload_postgres()

    # check the SSL settings
    cmd = [
        "openssl",
        "s_client",
        "-starttls",
        "postgres",
        "-connect",
        "172.27.1.2:5432",
        "-showcerts",
        "-CAfile",
        cluster.cert.crt,
    ]
    print(" ".join(cmd))
    p = subprocess.run(
        [
            "sudo",
            "-E",
            "-u",
            os.getenv("USER"),
            "env",
            "PATH=" + os.getenv("PATH"),
        ]
        + cmd,
        input="",
        text=True,
        capture_output=True,
    )
    if p.returncode != 0:
        print("" % p.stdout)
        print("" % p.stderr)
    assert p.returncode == 0

    # print connection string
    print("monitor: %s" % monitor.connection_string())
    monitor.check_ssl("on", "verify-ca")


def test_001_init_primary():
    global node1

    # Create a server certificate signed by the root Certificate Authority
    certs_dir = "/tmp/certs/node1"

    serverCert = cert.SSLCert(
        "/tmp/certs/node1", "server", "/CN=node1.pgautofailover.ca"
    )
    serverCert.create_signed_certificate(cluster.cert)

    # Now create the server with the certificates
    node1 = cluster.create_datanode(
        "/tmp/cert/node1",
        authMethod="skip",
        sslMode="verify-ca",
        sslCAFile=cluster.cert.crt,
        sslServerKey=serverCert.key,
        sslServerCert=serverCert.crt,
    )
    node1.create(level="-vv")

    with open(os.path.join("/tmp/cert/node1", "pg_hba.conf"), "a") as hba:
        # node1.run_sql_query will need
        # host "172.27.1.1", user "docker", database "postgres"
        hba.write("hostssl postgres docker %s cert\n" % cluster.networkSubnet)
        hba.write("hostssl all all %s cert\n" % cluster.networkSubnet)
        hba.write(
            "hostssl replication all %s cert map=pgautofailover\n"
            % cluster.networkSubnet
        )

    with open(os.path.join("/tmp/cert/node1", "pg_ident.conf"), "a") as ident:
        # use an ident map to allow using the same cert for replication
        ident.write("pgautofailover autoctl_node pgautofailover_replicator\n")

    node1.reload_postgres()

    node1.run()
    assert node1.wait_until_state(target_state="single")
    node1.wait_until_pg_is_running()
    node1.check_ssl("on", "verify-ca", primary=True)


def test_002_create_t1():
    print()
    print(node1.connection_string())
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_init_secondary():
    global node2

    # Create a server certificate signed by the root Certificate Authority
    certs_dir = "/tmp/certs/node2"

    serverCert = cert.SSLCert(
        "/tmp/certs/node2", "server", "/CN=node2.pgautofailover.ca"
    )
    serverCert.create_signed_certificate(cluster.cert)

    # Now create the server with the certificates
    node2 = cluster.create_datanode(
        "/tmp/cert/node2",
        authMethod="skip",
        sslMode="verify-ca",
        sslCAFile=cluster.cert.crt,
        sslServerKey=serverCert.key,
        sslServerCert=serverCert.crt,
    )
    node2.create(level="-vv")

    with open(os.path.join("/tmp/cert/node2", "pg_hba.conf"), "a") as hba:
        hba.write("hostssl all all %s cert\n" % cluster.networkSubnet)
        hba.write(
            "hostssl replication all %s cert map=pgautofailover\n"
            % cluster.networkSubnet
        )

    with open(os.path.join("/tmp/cert/node1", "pg_ident.conf"), "a") as ident:
        # use an ident map to allow using the same cert for replication
        ident.write("pgautofailover autoctl_node pgautofailover_replicator\n")

    node2.reload_postgres()

    node2.run()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")
    node2.wait_until_pg_is_running()
    node2.check_ssl("on", "verify-ca")


def test_004_failover():
    print()
    print("Calling pgautofailover.failover() on the monitor")
    cluster.monitor.failover()
    assert node2.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")



from nose.tools import *
from tests.tablespaces.tablespace_utils import node1
from tests.tablespaces.tablespace_utils import node2


# Failover is performed by the makefile Makefile, wait for it to complete
def test_001_wait_for_failover_and_insert():
    assert node2.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")

    node2.run_sql_query("INSERT INTO t2 VALUES (7)")
    results = node1.run_sql_query("SELECT * FROM t2")
    assert results == [(3,), (4,), (7,)]

    node2.run_sql_query("INSERT INTO t3 VALUES (8)")
    results = node1.run_sql_query("SELECT * FROM t3")
    assert results == [(5,), (6,), (8,)]



import time
from tests.pgautofailover_utils import QueryRunner
from tests.pgautofailover_utils import StatefulNode


class PGNodeTS(QueryRunner):
    def run_sql_query(self, query, *args):
        return super().run_sql_query(query, True, *args)

    def connection_string(self):
        return "postgresql://%s@%s:%s/%s" % (
            self.username,
            self.service_name,
            self.port,
            self.database,
        )


class MonitorNodeTS(PGNodeTS):
    def __init__(self, port, service_name):
        self.port = port
        self.service_name = service_name
        self.monitor = self
        self.username = "autoctl_node"
        self.database = "pg_auto_failover"


class DataNodeTS(PGNodeTS, StatefulNode):
    def __init__(self, port, service_name, monitor_node):
        self.port = port
        self.service_name = service_name
        self.monitor = monitor_node
        self.username = "docker"
        self.database = "postgres"

    def logger_name(self):
        return self.service_name

    def sleep(self, sleep_time):
        time.sleep(sleep_time)

    def print_debug_logs(self):
        # no-op, can't get logs easily in this test-type
        return

    def get_state(self):
        return super().get_state(
            "node with port %s not found on the monitor" % self.port,
            """
    SELECT reportedstate, goalstate
    FROM pgautofailover.node
    WHERE nodeport=%s
    """,
            self.port,
        )


monitor = MonitorNodeTS(
    "5432",
    "monitor",
)
node1 = DataNodeTS(
    "5433",
    "node1",
    monitor,
)
node2 = DataNodeTS(
    "5434",
    "node2",
    monitor,
)



import tests.pgautofailover_utils as pgautofailover
from nose.tools import raises, eq_
import time

import os.path

cluster = None
monitor = None
node1 = None
node2 = None
node3 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    global monitor
    monitor = cluster.create_monitor("/tmp/multi_ifdown/monitor")
    monitor.run()
    monitor.wait_until_pg_is_running()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/multi_ifdown/node1")
    node1.create()
    node1.run()
    assert node1.wait_until_state(target_state="single")


def test_002_add_standby():
    global node2

    node2 = cluster.create_datanode("/tmp/multi_ifdown/node2")
    node2.create()
    node2.run()

    assert node2.wait_until_pg_is_running()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()

    # make sure we reached primary on node1 before next tests
    assert node1.wait_until_state(target_state="primary")


def test_003_add_standby():
    global node3

    node3 = cluster.create_datanode("/tmp/multi_ifdown/node3")
    node3.create()
    node3.run()

    assert node3.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()
    assert node3.has_needed_replication_slots()

    # the formation number_sync_standbys is expected to be set to 1 now
    assert node1.get_number_sync_standbys() == 1

    # make sure we reached primary on node1 before next tests
    assert node1.wait_until_state(target_state="primary")


def test_004_write_into_primary():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2), (3), (4)")
    node1.run_sql_query("CHECKPOINT")

    results = node1.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,), (3,), (4,)]


def test_005_set_candidate_priorities():
    print()
    assert node1.wait_until_state(target_state="primary")

    # set priorities in a way that we know the candidate: node3
    node1.set_candidate_priority(90)  # current primary
    node2.set_candidate_priority(0)  # not a candidate anymore
    node3.set_candidate_priority(90)

    # when we set candidate priority we go to apply_settings then primary
    print()
    assert node1.wait_until_state(target_state="primary")
    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="secondary")

    # node1 should still be "sync"
    assert node1.get_number_sync_standbys() == 1
    assert node2.get_replication_quorum()

    # also let's see synchronous_standby_names here
    # remember to sort by candidate priority then name
    ssn = "ANY 1 (pgautofailover_standby_3, pgautofailover_standby_2)"
    node1.check_synchronous_standby_names(ssn)


def test_006_ifdown_node3():
    node3.ifdown()


def test_007_insert_rows():
    node1.run_sql_query(
        "INSERT INTO t1 SELECT x+10 FROM generate_series(1, 10000) as gs(x)"
    )
    node1.run_sql_query("CHECKPOINT")

    lsn1 = node1.run_sql_query("select pg_current_wal_lsn()")[0][0]
    print("%s " % lsn1, end="", flush=True)

    # node2 is sync and should get the WAL
    lsn2 = node2.run_sql_query("select pg_last_wal_receive_lsn()")[0][0]
    print("%s " % lsn2, end="", flush=True)

    while lsn2 != lsn1:
        time.sleep(1)
        lsn2 = node2.run_sql_query("select pg_last_wal_receive_lsn()")[0][0]
        print("%s " % lsn2, end="", flush=True)

    eq_(lsn1, lsn2)


def test_008_failover():
    print()
    print("Injecting failure of node1")
    node1.fail()

    # have node2 re-join the network and hopefully reconnect etc
    print("Reconnecting node3 (ifconfig up)")
    node3.ifup()

    # now we should be able to continue with the failover, and fetch missing
    # WAL bits from node2
    assert node3.wait_until_pg_is_running()

    assert node3.wait_until_state(target_state="wait_primary", timeout=120)
    assert node2.wait_until_state(target_state="secondary")

    # node 2 has candidate priority of 0, can still be used to reach primary
    assert node3.wait_until_state(target_state="primary")

    assert node3.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()

    # when in wait_primary state we should not block writes when:
    assert node3.get_number_sync_standbys() == 1

    ssn = "ANY 1 (pgautofailover_standby_1, pgautofailover_standby_2)"
    node3.check_synchronous_standby_names(ssn=ssn)


def test_009_read_from_new_primary():
    results = node3.run_sql_query("SELECT count(*) FROM t1")
    assert results == [(10004,)]


def test_010_start_node1_again():
    node1.run()
    assert node1.wait_until_state(target_state="secondary")

    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()
    assert node3.has_needed_replication_slots()

    # now that we're back to primary, check we have sync rep again
    ssn = "ANY 1 (pgautofailover_standby_1, pgautofailover_standby_2)"
    node3.check_synchronous_standby_names(ssn)


# test_011_XXX, test_012_XXX, test_013_XXX, test_014_XXX and test_015_XXX
# are meant to test the scenario when the most advanced secondary
# becomes inaccessible at the same time when the primary is inaccessible
def test_011_prepare_candidate_priorities():
    # we are aiming to promote node2
    assert node2.set_candidate_priority(100)  # the next primary

    # other nodes are already candidates for primary, but with less
    # priority
    assert node1.get_candidate_priority() == 90
    assert node3.get_candidate_priority() == 90


def test_012_prepare_replication_quorums():
    # for the purpose of this test, we need one node
    # async, to allow that we should decrement the sync stanbys
    node3.set_number_sync_standbys(0)

    # to emulate one node is behind, it is easier to make it async
    # we want node2 to be behind others
    assert node2.set_replication_quorum("false")

    # others should be sync
    assert node1.get_replication_quorum()
    assert node3.get_replication_quorum()


def test_013_secondary_gets_behind_primary():
    # make sure that node2 gets behind of the primary
    node2.ifdown()

    # primary ingests some data
    node3.run_sql_query("INSERT INTO t1 VALUES (5), (6)")
    node3.run_sql_query("CHECKPOINT")

    # ensure that the healthy secondary gets the change
    results = node1.run_sql_query("SELECT count(*) FROM t1")
    assert results == [(10006,)]

    lsn1 = node1.run_sql_query("select pg_last_wal_receive_lsn()")[0][0]
    print("%s " % lsn1, end="", flush=True)

    # ensure the monitor received this lsn
    node1.pg_autoctl.sighup()  # wake up from the 10s node_active delay
    time.sleep(1)

    q = "select reportedlsn from pgautofailover.node where nodeid = 1"
    lsn1m = monitor.run_sql_query(q)[0][0]
    print("%s " % lsn1m, end="", flush=True)

    retry = 0
    while lsn1 != lsn1m and retry < 3:
        time.sleep(1)
        lsn1m = monitor.run_sql_query(q)[0][0]
        print("%s " % lsn1m, end="", flush=True)

    eq_(lsn1, lsn1m)


def test_014_secondary_reports_lsn():
    # make the primary and mostAdvanced secondary inaccessible
    # and the candidate for failover as accessible
    # which means that node2 will not be able to fetch wal
    # and blocked until the other secondary is up
    assert node1.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="primary")

    node3.ifdown()  # primary
    node1.ifdown()  # most advanced standby
    node2.ifup()  # failover candidate

    print()
    print("Calling pgautofailover.failover() on the monitor")
    monitor.failover()

    # node2 reports its LSN while others are inaccessible
    assert node2.wait_until_state(target_state="report_lsn")


def test_015_finalize_failover_after_most_advanced_secondary_gets_back():
    # when they are accessible again, both should become
    # secondaries
    node1.ifup()  # old most advanced secondary, now secondary
    node3.ifup()  # old primary, now secondary

    assert node1.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="primary")

    results = node2.run_sql_query("SELECT count(*) FROM t1")
    eq_(results, [(10006,)])



import tests.pgautofailover_utils as pgautofailover
from nose.tools import *

import time
import os.path

cluster = None
node1 = None
node2 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    monitor = cluster.create_monitor("/tmp/ensure/monitor")
    monitor.run()
    monitor.wait_until_pg_is_running()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/ensure/node1")
    print()
    print("create node1")
    node1.create()
    print("stop postgres")
    node1.stop_postgres()
    print("run node1")
    node1.run()
    print("wait until Postgres is running")
    node1.wait_until_pg_is_running()
    assert node1.wait_until_state(target_state="single")


def test_002_create_t1():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_003_init_secondary():
    global node2
    node2 = cluster.create_datanode("/tmp/ensure/node2")
    node2.create()
    node2.stop_postgres()
    node2.run()
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


def test_004_demoted():
    print()
    node1.stop_postgres()
    node1.stop_pg_autoctl()
    # we need the pg_autoctl process to run to reach the state demoted,
    # otherwise the monitor assigns that state to node1 but we never reach
    # it
    print("stopped pg_autoctl and postgres, now waiting for 30s")
    node2.sleep(30)
    node1.run()
    # We must not wait for PG to run, since otherwise we might miss the demoted
    # state

    assert node1.wait_until_state(target_state="demoted")

    # ideally we should be able to check that we refrain from starting
    # postgres again before calling the transition function
    print("re-starting pg_autoctl on node1")
    assert node1.wait_until_state(target_state="secondary")


def test_005_inject_error_in_node2():
    assert node2.wait_until_state(target_state="primary")

    # break Postgres setup on the primary, and restart Postgres: then
    # Postgres keeps failing to start, and pg_autoctl still communicates
    # with the monitor, which should still orchestrate a failover.
    pgconf = os.path.join(node2.datadir, "postgresql.conf")

    with open(pgconf, "a+") as f:
        f.write("\n")
        f.write("shared_preload_libraries='wrong_extension'\n")

    node2.restart_postgres()

    # the first step is the promotion of the other node as the new primary:
    assert node1.wait_until_state("wait_primary")

    # then when the failover happens, the new primary postgresql.conf gets
    # copied over, and we get the nodes back to primary/secondary
    assert node2.wait_until_state("secondary")
    assert node1.wait_until_state("primary")



import tests.pgautofailover_utils as pgautofailover
from nose.tools import raises, eq_
import time

import os.path

cluster = None
monitor = None
node1 = None
node2 = None
node3 = None
node4 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    global monitor
    monitor = cluster.create_monitor("/tmp/multi_maintenance/monitor")
    monitor.run()
    monitor.wait_until_pg_is_running()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/multi_maintenance/node1")
    node1.create()
    node1.run()
    assert node1.wait_until_state(target_state="single")


def test_002_add_standby():
    global node2

    node2 = cluster.create_datanode("/tmp/multi_maintenance/node2")
    node2.create()
    node2.run()

    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()

    # make sure we reached primary on node1 before next tests
    assert node1.wait_until_state(target_state="primary")

    node1.check_synchronous_standby_names(
        ssn="ANY 1 (pgautofailover_standby_2)"
    )


def test_003_add_standby():
    global node3

    node3 = cluster.create_datanode("/tmp/multi_maintenance/node3")
    node3.create()
    node3.run()

    assert node3.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()
    assert node3.has_needed_replication_slots()

    # the formation number_sync_standbys is expected to be set to 1 now
    eq_(node1.get_number_sync_standbys(), 1)

    ssn = "ANY 1 (pgautofailover_standby_2, pgautofailover_standby_3)"
    node1.check_synchronous_standby_names(ssn)

    # make sure we reached primary on node1 before next tests
    assert node1.wait_until_state(target_state="primary")


def test_004_write_into_primary():
    node1.run_sql_query("CREATE TABLE t1(a int)")
    node1.run_sql_query("INSERT INTO t1 VALUES (1), (2), (3), (4)")
    node1.run_sql_query("CHECKPOINT")

    results = node1.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,), (3,), (4,)]


def test_005_set_candidate_priorities():
    print()
    assert node1.wait_until_state(target_state="primary")

    # set priorities in a way that we know the candidate: node3
    node1.set_candidate_priority(80)  # current primary
    node2.set_candidate_priority(70)  # remain secondary
    node3.set_candidate_priority(90)  # favorite for failover

    # when we set candidate priority we go to apply_settings then primary
    print()
    assert node1.wait_until_state(target_state="primary")
    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="secondary")

    # node2 should still be "sync"
    eq_(node2.get_replication_quorum(), True)

    # other replication settings should still be the same as before
    eq_(node1.get_number_sync_standbys(), 1)

    ssn = "ANY 1 (pgautofailover_standby_3, pgautofailover_standby_2)"
    node1.check_synchronous_standby_names(ssn)


def test_006a_maintenance_and_failover():
    print()
    print("Enabling maintenance on node2")
    node2.enable_maintenance()
    assert node2.wait_until_state(target_state="maintenance")
    node2.stop_postgres()

    # assigned and goal state must be the same
    assert node1.wait_until_state(target_state="primary")

    # ssn is not changed during maintenance operations
    ssn = "ANY 1 (pgautofailover_standby_3, pgautofailover_standby_2)"
    eq_(node1.get_synchronous_standby_names(), ssn)
    eq_(node1.get_synchronous_standby_names_local(), ssn)

    print("Calling pgautofailover.failover() on the monitor")
    monitor.failover()
    assert node3.wait_until_state(target_state="primary")
    assert node1.wait_until_state(target_state="secondary")

    # now that node3 is primary, synchronous_standby_names has changed
    ssn = "ANY 1 (pgautofailover_standby_1, pgautofailover_standby_2)"
    node3.check_synchronous_standby_names(ssn)

    print("Disabling maintenance on node2, should connect to the new primary")
    node2.disable_maintenance()

    # allow manual checking of primary_conninfo
    primary_conninfo_ipaddr = str(node3.vnode.address)
    print("current primary is node3 at %s" % primary_conninfo_ipaddr)

    if node2.pgmajor() < 12:
        fn = "recovery.conf"
    else:
        fn = "postgresql-auto-failover-standby.conf"

    fn = os.path.join(node2.datadir, fn)
    conf = open(fn).read()

    if primary_conninfo_ipaddr not in conf:
        raise Exception(
            "Primary ip address %s not found in %s:\n%s"
            % (primary_conninfo_ipaddr, fn, conf)
        )

    assert node1.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="primary")

    # ssn is not changed during maintenance operations
    node3.check_synchronous_standby_names(ssn)

    assert node1.has_needed_replication_slots()
    assert node2.has_needed_replication_slots()
    assert node3.has_needed_replication_slots()


def test_006b_read_from_new_primary():
    results = node3.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,), (3,), (4,)]


def test_007a_node1_to_maintenance():
    print()
    assert node3.wait_until_state(target_state="primary")

    print("Enabling maintenance on node1")
    node1.enable_maintenance()

    assert node3.wait_until_state(target_state="primary")


def test_007b_node2_to_maintenance():
    # node3 is the current primary
    assert node3.get_number_sync_standbys() == 1

    print("Enabling maintenance on node2")
    node2.enable_maintenance()

    assert node3.wait_until_state(target_state="primary")

    # when both secondaries are put to maintenance, writes are blocked on
    # the primary
    ssn = "ANY 1 (pgautofailover_standby_1, pgautofailover_standby_2)"
    node3.check_synchronous_standby_names(ssn)


def test_008a_stop_primary():
    # node3 is the current primary
    assert node3.get_state().assigned == "primary"
    node3.fail()

    # check that even after 30s node3 is still not set to draining
    node3.sleep(30)
    assert not node3.get_state().assigned == "draining"
    assert node3.get_state().assigned == "primary"


def test_008b_start_primary():
    node3.run()
    assert node3.wait_until_state(target_state="primary")


@raises(Exception)
def test_009a_enable_maintenance_on_primary_should_fail():
    node3.enable_maintenance(allowFailover=True)


def test_009b_disable_maintenance():
    print("Disabling maintenance on node1 and node2")
    node1.disable_maintenance()
    node2.disable_maintenance()

    assert node1.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="primary")


def test_010_set_number_sync_standby_to_zero():
    assert node3.set_number_sync_standbys(0)
    eq_(node3.get_number_sync_standbys(), 0)


def test_011_all_to_maintenance():
    print()
    assert node3.wait_until_state(target_state="primary")

    print("Enabling maintenance on node1")
    node1.enable_maintenance()

    assert node3.wait_until_state(target_state="primary")

    # now we can, because we don't care about having any standbys
    print("Enabling maintenance on node2")
    node2.enable_maintenance()

    assert node3.wait_until_state(target_state="wait_primary")

    # also let's see synchronous_standby_names here
    node3.check_synchronous_standby_names(ssn="")


def test_012_can_write_during_maintenance():
    node3.run_sql_query("INSERT INTO t1 VALUES (5), (6)")
    node3.run_sql_query("CHECKPOINT")


def test_013_add_standby():
    global node4

    node4 = cluster.create_datanode("/tmp/multi_maintenance/node4")
    node4.create()
    node4.run()

    assert node4.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="primary")
    assert node2.wait_until_state(target_state="maintenance")
    assert node1.wait_until_state(target_state="maintenance")

    assert node3.has_needed_replication_slots()
    assert node4.has_needed_replication_slots()

    # the formation number_sync_standbys is expected to not be changed
    eq_(node3.get_number_sync_standbys(), 0)

    # make sure we reached primary on node1 before next tests
    assert node3.wait_until_state(target_state="primary")


def test_014_disable_maintenance():
    print()
    # make sure node2 is still in maintenance, then disable maintenance
    print("Disabling maintenance on node2")
    assert node2.wait_until_state(target_state="maintenance")
    node2.disable_maintenance()

    assert node3.wait_until_state(target_state="primary")

    print("Disabling maintenance on node1")
    # make sure node1 is still in maintenance, then disable maintenance
    assert node1.wait_until_state(target_state="maintenance")
    node1.disable_maintenance()

    assert node3.wait_until_state(target_state="primary")

    # also let's see synchronous_standby_names here
    print("Monitor: %s" % node3.get_synchronous_standby_names())
    print(
        "Node 3:  %s"
        % node3.run_sql_query("show synchronous_standby_names")[0][0]
    )


def test_015_set_number_sync_standby_to_one():
    node3.set_number_sync_standbys(1)
    eq_(node3.get_number_sync_standbys(), 1)

    assert node3.wait_until_state(target_state="primary")


def test_016_two_standbys_in_maintenance():
    print()

    print("Enabling maintenance on node1")
    node1.enable_maintenance()

    assert node3.wait_until_state(target_state="primary")

    print("Enabling maintenance on node2")
    node2.enable_maintenance()

    assert node3.wait_until_state(target_state="primary")


@raises(Exception)
def test_017_primary_to_maintenance():
    # this should fail, because we have 4 nodes and number_sync_standbys = 1
    # and 2 nodes are already in maintenance
    #
    # if we allowed a 3rd node to be in maintenance, we would have no
    # standby node left and 0 < 1
    print()
    print("Enabling maintenance on node3 (primary)")
    node3.enable_maintenance()


def test_018_disable_maintenance():
    print()
    print("Disabling maintenance on node2")
    assert node2.wait_until_state(target_state="maintenance")
    node2.disable_maintenance()

    assert node3.wait_until_state(target_state="primary")

    # make sure node1 is still in maintenance, then disable maintenance
    print("Disabling maintenance on node1")
    assert node1.wait_until_state(target_state="maintenance")
    node1.disable_maintenance()

    assert node3.wait_until_state(target_state="primary")


def test_019_set_priorities():
    # set priorities in a way that we know the candidate: node1
    node1.set_candidate_priority(90)
    node2.set_candidate_priority(70)
    node3.set_candidate_priority(70)  # current primary
    node4.set_candidate_priority(70)

    # when we set candidate priority we go to apply_settings then primary
    print()
    assert node1.wait_until_state(target_state="secondary")
    assert node2.wait_until_state(target_state="secondary")
    assert node3.wait_until_state(target_state="primary")
    assert node4.wait_until_state(target_state="secondary")


def test_020_primary_to_maintenance():
    print()
    assert node3.wait_until_state(target_state="primary")

    print("Enabling maintenance on node3, allowing failover")
    node3.enable_maintenance(allowFailover=True)

    assert node3.wait_until_state(target_state="maintenance")
    assert node2.wait_until_state(target_state="secondary")
    assert node4.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


def test_021_stop_maintenance():
    print()
    print("Disabling maintenance on node3")
    node3.disable_maintenance()

    assert node3.wait_until_pg_is_running()
    assert node3.wait_until_state(target_state="secondary")

    assert node1.wait_until_state(target_state="primary")

    assert node2.wait_until_state(target_state="secondary")
    assert node4.wait_until_state(target_state="secondary")



import tests.pgautofailover_utils as pgautofailover
import psycopg2
from nose.tools import *

ha_cluster = None
monitor = None
coordinator1a = None
coordinator1b = None
# we will be creating a citus cluster with 2 workers
# each has a secondary for replication. Workers are numbered
# a/b indicates primary/secondary (only true before first failover)
# a/b swap around primary and secondary roles when failures are induced
worker1a = None
worker1b = None
worker2a = None
worker2b = None


def setup_module():
    global ha_cluster
    ha_cluster = pgautofailover.Cluster()


def teardown_module():
    coordinator1b.run_sql_query("select public.wait_until_metadata_sync()")
    coordinator1b.run_sql_query("DROP TABLE t1")
    ha_cluster.destroy()


def test_000_create_monitor():
    global monitor
    monitor = ha_cluster.create_monitor("/tmp/citus/force/monitor")
    monitor.run()


def test_001_init_coordinator():
    global coordinator1a
    global coordinator1b

    coordinator1a = ha_cluster.create_datanode(
        "/tmp/citus/force/coordinator1a", role=pgautofailover.Role.Coordinator
    )
    coordinator1a.create(name="coord1a")
    coordinator1a.run()

    print()  # make the debug output more readable
    assert coordinator1a.wait_until_state(target_state="single")

    # we need to expose some Citus testing internals
    assert coordinator1a.wait_until_pg_is_running()
    coordinator1a.create_wait_until_metadata_sync()

    coordinator1b = ha_cluster.create_datanode(
        "/tmp/citus/force/coordinator1b", role=pgautofailover.Role.Coordinator
    )
    coordinator1b.create(name="coord1b")
    coordinator1b.run()
    assert coordinator1a.wait_until_state(target_state="primary")
    assert coordinator1b.wait_until_state(target_state="secondary")


def test_002_init_workers():
    global worker1a
    global worker1b
    global worker2a
    global worker2b

    worker1a = ha_cluster.create_datanode(
        "/tmp/citus/force/worker1a", group=1, role=pgautofailover.Role.Worker
    )
    worker1a.create(name="worker1a")
    worker1a.run()

    worker1b = ha_cluster.create_datanode(
        "/tmp/citus/force/worker1b", group=1, role=pgautofailover.Role.Worker
    )
    worker1b.create(name="worker1b")
    worker1b.run()

    worker2a = ha_cluster.create_datanode(
        "/tmp/citus/force/worker2a", group=2, role=pgautofailover.Role.Worker
    )
    worker2a.create(name="worker2a")
    worker2a.run()

    worker2b = ha_cluster.create_datanode(
        "/tmp/citus/force/worker2b", group=2, role=pgautofailover.Role.Worker
    )
    worker2b.create(name="worker2b")
    worker2b.run()

    # wait here till all workers are stable and in the desired state
    # by not waiting on all workers separately this should save a bit of time
    print()  # make the debug output more readable
    assert worker1a.wait_until_state(target_state="primary")
    assert worker1b.wait_until_state(target_state="secondary")
    assert worker2a.wait_until_state(target_state="primary")
    assert worker2b.wait_until_state(target_state="secondary")


def test_003_create_distributed_table():
    coordinator1a.run_sql_query(
        """
    SET citus.shard_count TO 4;
    CREATE TABLE t1 (a int);
    SELECT create_distributed_table('t1', 'a');
    """
    )


def test_004_fail_while_transaction_is_in_progress():
    # can't use with here :( when using that it will throw an exception once the
    # connection gets closed due to the failover killing the competing backend
    conn = psycopg2.connect(coordinator1a.connection_string())
    cur = conn.cursor()
    # this becomes a multi worker query with locks held on the shards
    # these locks compete with the failover that will be triggered by worker1a.fail
    # if no progress is made on this transaction the failover stays blocked.
    # This is undesirable so the failover mechanism waits for a certain time before
    # aggressively terminating backends that keep competing locks that would prevent the
    # failvover from completing.
    cur.execute("INSERT INTO t1 VALUES (1), (2)")  # has nothing to fetch

    worker1a.fail()
    print()  # make the debug output more readable
    assert worker1b.wait_until_state(target_state="wait_primary")

    try:
        # should throw exception as the connection has been terminated
        cur.execute("INSERT INTO t1 VALUES (3)")
        raise Exception(
            "competing transaction should have been terminated during failover"
        )
    except psycopg2.OperationalError:
        # expected behaviour
        pass

    conn.close()


def test_005_drop_primary_worker():
    worker2a.fail()
    # can't drop a worker node with shard placements, wait until failover
    worker2b.wait_until_state("wait_primary")
    worker2a.drop()
    worker2b.wait_until_state("single")


def test_005_drop_primary_coordinator():
    coordinator1a.drop()
    coordinator1b.wait_until_state("single")



import tests.pgautofailover_utils as pgautofailover
from nose.tools import assert_raises, raises, eq_

import os
import shutil
import subprocess
import time

cluster = None
monitor = None
node1 = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    cluster.destroy()


def test_000_create_monitor():
    global monitor
    monitor = cluster.create_monitor("/tmp/config_test/monitor")
    monitor.run()


def test_001_init_primary():
    global node1
    node1 = cluster.create_datanode("/tmp/config_test/node1")
    node1.create()

    # the name of the node should be "%s_%d" % ("node", node1.nodeid)
    eq_(node1.get_nodename(), "node_%d" % node1.get_nodeid())

    # we can change the name on the monitor with pg_autoctl set node metadata
    node1.set_metadata(name="node a")
    eq_(node1.get_nodename(), "node a")

    node1.run()
    assert node1.wait_until_state(target_state="single")

    # we can also change the name directly in the configuration file
    node1.config_set("pg_autoctl.name", "a")

    # wait until the reload signal has been processed before checking
    time.sleep(2)
    eq_(node1.get_nodename(), "a")


def test_002_config_set_monitor():
    pg_ctl = monitor.config_get("postgresql.pg_ctl")

    # set something non-default to assert no side-effects later
    sslmode = "prefer"
    monitor.config_set("ssl.sslmode", sslmode)

    # set monitor config postgresql.pg_ctl to something invalid
    with assert_raises(subprocess.CalledProcessError):
        monitor.config_set("postgresql.pg_ctl", "invalid")

    # it should not get changed
    eq_(monitor.config_get("postgresql.pg_ctl"), pg_ctl)

    # try again with a keeper
    pg_ctl = node1.config_get("postgresql.pg_ctl")

    # set the keeper to something invalid
    with assert_raises(subprocess.CalledProcessError):
        node1.config_set("postgresql.pg_ctl", "invalid")

    # it should not get changed
    eq_(node1.config_get("postgresql.pg_ctl"), pg_ctl)

    # pg_ctl can be moved and `config set` will still operate.
    shutil.copy(pg_ctl, "/tmp/pg_ctl")
    monitor.config_set("postgresql.pg_ctl", "/tmp/pg_ctl")
    # "move" pg_ctl
    os.remove("/tmp/pg_ctl")
    monitor.config_set("postgresql.pg_ctl", pg_ctl)

    eq_(monitor.config_get("postgresql.pg_ctl"), pg_ctl)

    # no side effects
    eq_(monitor.config_get("ssl.sslmode"), sslmode)



from nose.tools import *
from tests.tablespaces.tablespace_utils import node1
from tests.tablespaces.tablespace_utils import node2


# Node 2 is spun up by Makefile, wait until it's up and replicating
def test_001_init_secondary():
    assert node2.wait_until_state(target_state="secondary")
    assert node1.wait_until_state(target_state="primary")


def test_002_read_from_secondary():
    results = node2.run_sql_query("SELECT * FROM t1")
    assert results == [(1,), (2,)]
    results = node2.run_sql_query("SELECT * FROM t2")
    assert results == [(3,), (4,)]


def test_003_create_tablespace_while_streaming():
    node1.run_sql_query(
        "CREATE TABLESPACE extended_b LOCATION '/extra_volumes/extended_b';"
    )
    node1.run_sql_query("CREATE TABLE t3(i int) TABLESPACE extended_b;")
    node1.run_sql_query("INSERT INTO t3 VALUES (5), (6)")


def test_004_read_from_secondary_again():
    results = node2.run_sql_query("SELECT * FROM t3")
    assert results == [(5,), (6,)]



import tests.pgautofailover_utils as pgautofailover
from nose.tools import eq_, raises

import os.path
import time
import subprocess
import pprint

cluster = None
monitor = None
coordinator1a = None
coordinator1b = None
# we will be creating a citus cluster with 2 workers
# each has a secondary for replication. Workers are numbered
# a/b indicates primary/secondary (only true before first failover)
# a/b swap around primary and secondary roles when failures are induced
worker1a = None
worker1b = None
worker2a = None
worker2b = None


def setup_module():
    global cluster
    cluster = pgautofailover.Cluster()


def teardown_module():
    coordinator1b.run_sql_query("select public.wait_until_metadata_sync()")
    coordinator1b.run_sql_query("DROP TABLE t1")
    cluster.destroy()


def test_000_create_monitor():
    global monitor
    monitor = cluster.create_monitor("/tmp/citus/basic/monitor")
    monitor.run()


def test_001_init_coordinator():
    global coordinator1a
    global coordinator1b

    coordinator1a = cluster.create_datanode(
        "/tmp/citus/basic/coordinator1a", role=pgautofailover.Role.Coordinator
    )
    coordinator1a.create()
    coordinator1a.run()

    assert coordinator1a.wait_until_state(target_state="single")

    # we need to expose some Citus testing internals
    assert coordinator1a.wait_until_pg_is_running()
    coordinator1a.create_wait_until_metadata_sync()

    # check that Citus did skip creating the SSL certificates (--no-ssl)
    assert not os.path.isfile(
        os.path.join("/tmp/citus/basic/coordinator1a", "server.key")
    )

    coordinator1b = cluster.create_datanode(
        "/tmp/citus/basic/coordinator1b", role=pgautofailover.Role.Coordinator
    )
    coordinator1b.create()
    coordinator1b.run()
    assert coordinator1a.wait_until_state(target_state="primary")
    assert coordinator1b.wait_until_state(target_state="secondary")

    sql = "select nodename, nodeport from pg_dist_node"
    nodes = coordinator1a.run_sql_query(sql)

    assert nodes[0][0] == str(coordinator1a.vnode.address)
    assert nodes[0][1] == coordinator1a.port


def test_002_init_workers():
    global worker1a
    global worker1b
    global worker2a
    global worker2b

    worker1a = cluster.create_datanode(
        "/tmp/citus/basic/worker1a", group=1, role=pgautofailover.Role.Worker
    )
    worker1a.create(name="worker1a")
    worker1a.run()

    worker1b = cluster.create_datanode(
        "/tmp/citus/basic/worker1b", group=1, role=pgautofailover.Role.Worker
    )
    worker1b.create(name="worker1b")
    worker1b.run()

    worker2a = cluster.create_datanode(
        "/tmp/citus/basic/worker2a", group=2, role=pgautofailover.Role.Worker
    )
    worker2a.create(name="worker2a")
    worker2a.run()

    worker2b = cluster.create_datanode(
        "/tmp/citus/basic/worker2b", group=2, role=pgautofailover.Role.Worker
    )
    worker2b.create(name="worker2b")
    worker2b.run()

    # wait here till all workers are stable and in the desired state
    # by not waiting on all workers separately this should save a bit of time
    print()  # make the debug output more readable
    assert worker1a.wait_until_state(target_state="primary")
    assert worker1b.wait_until_state(target_state="secondary")
    assert worker2a.wait_until_state(target_state="primary")
    assert worker2b.wait_until_state(target_state="secondary")


def test_003_create_distributed_table():
    coordinator1a.run_sql_query("CREATE TABLE t1 (a int)")
    coordinator1a.run_sql_query("SELECT create_distributed_table('t1', 'a')")
    coordinator1a.run_sql_query("INSERT INTO t1 VALUES (1), (2)")


def test_004_001_fail_worker2():
    worker2a.fail()


@raises(Exception)
def test_004_002_writes_via_coordinator_to_worker2_fail():
    # value 3 is routed to the worker2 pair, which we just failed and
    # didn't had time to fail over yet. This will give an error due
    # to the failure of citus to contact the worker that just failed
    coordinator1a.run_sql_query("INSERT INTO t1 VALUES (3)")


def test_004_003_reads_for_worker1_via_coordinator_work():
    # value 2 is routed to the worker1 pair. Since there is no failure
    # in this pair the coordinator is able to read the information
    results = coordinator1a.run_sql_query("SELECT a FROM t1 WHERE a = 2")
    eq_(results, [(2,)])


def test_004_004_wait_for_failover():
    print()  # make the debug output more readable
    assert worker2b.wait_until_state(target_state="wait_primary")


def test_004_005_writes_via_coordinator_to_worker2_succeed_after_failover():
    # same as test 004_002, but now that the failover has been performed
    # the value can be correctly written to worker2b (worker2a is still dead)
    coordinator1a.run_sql_query("INSERT INTO t1 VALUES (3)")


def test_005_read_from_workers_via_coordinator():
    results = coordinator1a.run_sql_query("SELECT a FROM t1 ORDER BY a ASC")
    eq_(results, [(1,), (2,), (3,)])


def test_006_writes_to_coordinator_succeed():
    coordinator1a.run_sql_query("INSERT INTO t1 VALUES (4)")
    results = coordinator1a.run_sql_query("SELECT a FROM t1 ORDER BY a ASC")
    eq_(results, [(1,), (2,), (3,), (4,)])


def test_007_start_worker2_again():
    worker2a.run()

    print()  # make the debug output more readable
    assert worker2b.wait_until_state(target_state="primary")
    assert worker2a.wait_until_state(target_state="secondary")

    ssn = worker2b.run_sql_query("show synchronous_standby_names")[0][0]
    eq_(ssn, "ANY 1 (pgautofailover_standby_5)")


def test_008_read_from_workers_via_coordinator():
    results = coordinator1a.run_sql_query("SELECT a FROM t1 ORDER BY a ASC")
    eq_(results, [(1,), (2,), (3,), (4,)])


def test_009_fail_worker2b():
    worker2b.fail()

    print()  # make the debug output more readable
    assert worker2a.wait_until_state(target_state="wait_primary")


def test_010_read_from_workers_via_coordinator():
    # after failing back to worker2a verify the value written to worker2b
    # has been safely replicated to worker2a
    results = coordinator1a.run_sql_query("SELECT a FROM t1 ORDER BY a ASC")
    eq_(results, [(1,), (2,), (3,), (4,)])


def test_011_start_worker2b_again():
    worker2b.run()

    print()
    assert worker2b.wait_until_state(target_state="secondary")
    assert worker2a.wait_until_state(target_state="primary")


def test_012_perform_failover_worker2():
    print()

    print("Calling pgautofailover.failover(group => 2) on the monitor")
    cluster.monitor.failover(group=2)
    assert worker2a.wait_until_state(target_state="secondary")
    assert worker2b.wait_until_state(target_state="primary")

    assert worker2a.has_needed_replication_slots()
    assert worker2b.has_needed_replication_slots()


# we had a bug where if the old primary has already reached draining (both
# assigned and reported the state) when the secondary starts the transition
# from secondary to prepare_promotion, then the secondary transition fails
# because calling pgautofailover.get_primary() on the monitor returned no
# rows.
#
# that's fixed by not calling pgautofailover.get_primary() in the
# transition, it can be avoided easily.
#
def test_013_perform_failover_worker2b_draining():
    print()

    worker2a.stop_pg_autoctl()

    print("Calling pgautofailover.failover(group => 2) on the monitor")
    cluster.monitor.failover(group=2)
    assert worker2b.wait_until_state(target_state="draining")

    worker2a.run()

    assert worker2b.wait_until_state(target_state="secondary")
    assert worker2a.wait_until_state(target_state="primary")

    assert worker2a.has_needed_replication_slots()
    assert worker2b.has_needed_replication_slots()


def test_014_perform_failover_coordinator():
    print()

    print("Calling pgautofailover.failover(group => 0) on the monitor")
    cluster.monitor.failover(group=0)
    assert coordinator1a.wait_until_state(target_state="secondary")
    assert coordinator1b.wait_until_state(target_state="primary")

    assert coordinator1a.has_needed_replication_slots()
    assert coordinator1b.has_needed_replication_slots()
